##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Base class for control volumes
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, Param, Reals, Var
from pyomo.dae import DerivativeVar

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        ControlVolumeBlockData,
                        FlowDirection,
                        MaterialFlowBasis)
from idaes.core.util.exceptions import (BalanceTypeNotSupportedError,
                                        ConfigurationError,
                                        PropertyNotSupportedError,
                                        PropertyPackageError)
from idaes.core.util.tables import create_stream_table_dataframe

from idaes.logger import getIdaesLogger, getInitLogger, init_tee

__author__ = "Andrew Lee"


_log = getIdaesLogger(__name__)

# TODO : Custom terms in material balances, other types of material balances
# TODO : Improve flexibility for get_material_flow_terms and associated


@declare_process_block_class("ControlVolume0DBlock", doc="""
    ControlVolume0DBlock is a specialized Pyomo block for IDAES non-discretized
    control volume blocks, and contains instances of ControlVolume0DBlockData.

    ControlVolume0DBlock should be used for any control volume with a defined
    volume and distinct inlets and outlets which does not require spatial
    discretization. This encompases most basic unit models used in process
    modeling.""")
class ControlVolume0DBlockData(ControlVolumeBlockData):
    """
    0-Dimensional (Non-Discretised) ControlVolume Class

    This class forms the core of all non-discretized IDAES models. It provides
    methods to build property and reaction blocks, and add mass, energy and
    momentum balances. The form of the terms used in these constraints is
    specified in the chosen property package.
    """
    def build(self):
        """
        Build method for ControlVolume0DBlock blocks.

        Returns:
            None
        """
        # Call build method from base class
        super(ControlVolume0DBlockData, self).build()

    def add_geometry(self):
        """
        Method to create volume Var in ControlVolume.

        Args:
            None

        Returns:
            None
        """
        l_units = self.config.property_package.get_metadata().default_units[
                                                                      "length"]
        self.volume = Var(self.flowsheet().config.time, initialize=1.0,
                          doc='Holdup Volume [{}^3]'.format(l_units))

    def add_state_blocks(self,
                         information_flow=FlowDirection.forward,
                         has_phase_equilibrium=None):
        """
        This method constructs the inlet and outlet state blocks for the
        control volume.

        Args:
            information_flow: a FlowDirection Enum indicating whether
                               information flows from inlet-to-outlet or
                               outlet-to-inlet
            has_phase_equilibrium: indicates whether equilibrium calculations
                                    will be required in state blocks
            package_arguments: dict-like object of arguments to be passed to
                                state blocks as construction arguments
        Returns:
            None
        """
        if has_phase_equilibrium is None:
            raise ConfigurationError(
                    "{} add_state_blocks method was not provided with a "
                    "has_phase_equilibrium argument.".format(self.name))
        elif has_phase_equilibrium not in [True, False]:
            raise ConfigurationError(
                    "{} add_state_blocks method was provided with an invalid "
                    "has_phase_equilibrium argument. Must be True or False"
                    .format(self.name))

        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = has_phase_equilibrium
        tmp_dict["parameters"] = self.config.property_package

        if information_flow == FlowDirection.forward:
            tmp_dict["defined_state"] = True
        elif information_flow == FlowDirection.backward:
            tmp_dict["defined_state"] = False
        else:
            raise ConfigurationError(
                    '{} invalid value for information_flow argument. '
                    'Valid values are FlowDirection.forward and '
                    'FlowDirection.backward'.format(self.name))

        try:
            self.properties_in = (
                    self.config.property_package.state_block_class(
                            self.flowsheet().config.time,
                            doc="Material properties at inlet",
                            default=tmp_dict))

            # Reverse defined_state
            tmp_dict_2 = dict(**tmp_dict)
            tmp_dict_2["defined_state"] = not tmp_dict["defined_state"]

            self.properties_out = (
                    self.config.property_package.state_block_class(
                            self.flowsheet().config.time,
                            doc="Material properties at outlet",
                            default=tmp_dict_2))
        except AttributeError:
            raise PropertyPackageError(
                    "{} physical property package has not implemented the "
                    "state_block_class attribute. Please contact the "
                    "developer of the physical property package."
                    .format(self.name))

    def add_reaction_blocks(self, has_equilibrium=None):
        """
        This method constructs the reaction block for the control volume.

        Args:
            has_equilibrium: indicates whether equilibrium calculations
                              will be required in reaction block
            package_arguments: dict-like object of arguments to be passed to
                                reaction block as construction arguments

        Returns:
            None
        """
        if has_equilibrium is None:
            raise ConfigurationError(
                    "{} add_reaction_blocks method was not provided with a "
                    "has_equilibrium argument.".format(self.name))
        elif has_equilibrium not in [True, False]:
            raise ConfigurationError(
                    "{} add_reaction_blocks method was provided with an "
                    "invalid has_equilibrium argument. Must be True or False"
                    .format(self.name))

        tmp_dict = dict(**self.config.reaction_package_args)
        tmp_dict["state_block"] = self.properties_out
        tmp_dict["has_equilibrium"] = has_equilibrium
        tmp_dict["parameters"] = self.config.reaction_package

        try:
            self.reactions = (
                    self.config.reaction_package.reaction_block_class(
                            self.flowsheet().config.time,
                            doc="Reaction properties in control volume",
                            default=tmp_dict))
        except AttributeError:
            raise PropertyPackageError(
                    "{} reaction property package has not implemented the "
                    "reaction_block_class attribute. Please contact the "
                    "developer of the reaction property package."
                    .format(self.name))

    def add_phase_component_balances(self,
                                     has_rate_reactions=False,
                                     has_equilibrium_reactions=False,
                                     has_phase_equilibrium=False,
                                     has_mass_transfer=False,
                                     custom_molar_term=None,
                                     custom_mass_term=None):
        """
        This method constructs a set of 0D material balances indexed by time,
        phase and component.

        Args:
            has_rate_reactions: whether default generation terms for rate
                    reactions should be included in material balances
            has_equilibrium_reactions: whether generation terms should for
                    chemical equilibrium reactions should be included in
                    material balances
            has_phase_equilibrium: whether generation terms should for phase
                    equilibrium behaviour should be included in material
                    balances
            has_mass_transfer: whether generic mass transfer terms should be
                    included in material balances
            custom_molar_term: a Pyomo Expression representing custom terms to
                    be included in material balances on a molar basis.
                    Expression must be indexed by time, phase list and
                    component list
            custom_mass_term: a Pyomo Expression representing custom terms to
                    be included in material balances on a mass basis.
                    Expression must be indexed by time, phase list and
                    component list

        Returns:
            Constraint object representing material balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        # Check that reaction block exists if required
        if has_rate_reactions or has_equilibrium_reactions:
            try:
                rblock = self.reactions
            except AttributeError:
                raise ConfigurationError(
                        "{} does not contain a Reaction Block, but material "
                        "balances have been set to contain reaction terms. "
                        "Please construct a reaction block before adding "
                        "balance equations.".format(self.name))

        if has_equilibrium_reactions:
            # Check that reaction block is set to calculate equilibrium
            for t in self.flowsheet().config.time:
                if self.reactions[t].config.has_equilibrium is False:
                    raise ConfigurationError(
                            "{} material balance was set to include "
                            "equilibrium reactions, however the associated "
                            "ReactionBlock was not set to include equilibrium "
                            "constraints (has_equilibrium_reactions=False). "
                            "Please correct your configuration arguments."
                            .format(self.name))

        if has_phase_equilibrium:
            # Check that state blocks are set to calculate equilibrium
            for t in self.flowsheet().config.time:
                if not self.properties_out[t].config.has_phase_equilibrium:
                    raise ConfigurationError(
                            "{} material balance was set to include phase "
                            "equilibrium, however the associated outlet "
                            "StateBlock was not set to include equilibrium "
                            "constraints (has_phase_equilibrium=False). Please"
                            " correct your configuration arguments."
                            .format(self.name))
                if not self.properties_in[t].config.has_phase_equilibrium:
                    raise ConfigurationError(
                            "{} material balance was set to include phase "
                            "equilibrium, however the associated inlet "
                            "StateBlock was not set to include equilibrium "
                            "constraints (has_phase_equilibrium=False). Please"
                            " correct your configuration arguments."
                            .format(self.name))

        # Get units from property package
        units = {}
        for u in ['length', 'holdup', 'amount', 'time']:
            try:
                units[u] = \
                   self.config.property_package.get_metadata().default_units[u]
            except KeyError:
                units[u] = '-'

        # Test for components that must exist prior to calling this method
        if has_holdup:
            if not hasattr(self, "volume"):
                raise ConfigurationError(
                        "{} control volume must have volume defined to have "
                        "holdup and/or rate reaction terms. Please call the "
                        "add_geometry method before adding balance equations."
                        .format(self.name))

        # Material holdup and accumulation
        if has_holdup:
            self.material_holdup = Var(self.flowsheet().config.time,
                                       self.config.property_package.phase_list,
                                       self.config.property_package.
                                       component_list,
                                       domain=Reals,
                                       initialize=1.0,
                                       doc="Material holdup in unit [{}]"
                                           .format(units['holdup']))
        if dynamic:
            self.material_accumulation = DerivativeVar(
                    self.material_holdup,
                    wrt=self.flowsheet().config.time,
                    doc="Material accumulation in unit [{}/{}]"
                        .format(units['holdup'], units['time']))

        # Get phase component list(s)
        phase_component_list = self._get_phase_comp_list()

        # Create material balance terms as required
        # Kinetic reaction generation
        if has_rate_reactions:
            if not hasattr(self.config.reaction_package, "rate_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of rate "
                    "reactions (rate_reaction_idx), thus does not support "
                    "rate-based reactions.".format(self.name))
            self.rate_reaction_generation = Var(
                        self.flowsheet().config.time,
                        self.config.property_package.phase_list,
                        self.config.property_package.component_list,
                        domain=Reals,
                        initialize=0.0,
                        doc="Amount of component generated in "
                            "unit by kinetic reactions [{}/{}]"
                            .format(units['holdup'], units['time']))

        # Equilibrium reaction generation
        if has_equilibrium_reactions:
            if not hasattr(self.config.reaction_package,
                           "equilibrium_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of "
                    "equilibrium reactions (equilibrium_reaction_idx), thus "
                    "does not support equilibrium-based reactions."
                    .format(self.name))
            self.equilibrium_reaction_generation = Var(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated in unit "
                    "by equilibrium reactions [{}/{}]"
                    .format(units['holdup'], units['time']))

        # Phase equilibrium generation
        if has_phase_equilibrium:
            if not hasattr(self.config.property_package,
                           "phase_equilibrium_idx"):
                raise PropertyNotSupportedError(
                    "{} Property package does not contain a list of phase "
                    "equilibrium reactions (phase_equilibrium_idx), thus does "
                    "not support phase equilibrium.".format(self.name))
            self.phase_equilibrium_generation = Var(
                self.flowsheet().config.time,
                self.config.property_package.phase_equilibrium_idx,
                domain=Reals,
                initialize=0.0,
                doc="Amount of generation in unit by phase "
                    "equilibria [{}/{}]"
                    .format(units['holdup'], units['time']))

        # Material transfer term
        if has_mass_transfer:
            self.mass_transfer_term = Var(
                        self.flowsheet().config.time,
                        self.config.property_package.phase_list,
                        self.config.property_package.component_list,
                        domain=Reals,
                        initialize=0.0,
                        doc="Component material transfer into unit [{}/{}]"
                            .format(units['holdup'], units['time']))

        # Create rules to substitute material balance terms
        # Accumulation term
        def accumulation_term(b, t, p, j):
            return b.material_accumulation[t, p, j] if dynamic else 0

        def kinetic_term(b, t, p, j):
            return (b.rate_reaction_generation[t, p, j] if has_rate_reactions
                    else 0)

        def equilibrium_term(b, t, p, j):
            return (b.equilibrium_reaction_generation[t, p, j]
                    if has_equilibrium_reactions else 0)

        def phase_equilibrium_term(b, t, p, j):
            if has_phase_equilibrium:
                sd = {}
                for r in b.config.property_package.phase_equilibrium_idx:
                    if b.config.property_package.\
                            phase_equilibrium_list[r][0] == j:
                        if b.config.property_package.\
                                phase_equilibrium_list[r][1][0] == p:
                            sd[r] = 1
                        elif b.config.property_package.\
                                phase_equilibrium_list[r][1][1] == p:
                            sd[r] = -1
                        else:
                            sd[r] = 0
                    else:
                        sd[r] = 0

                return sum(b.phase_equilibrium_generation[t, r] * sd[r]
                           for r in b.config.property_package.
                           phase_equilibrium_idx)
            else:
                return 0

        def transfer_term(b, t, p, j):
            return (b.mass_transfer_term[t, p, j] if has_mass_transfer else 0)

        def user_term_mol(b, t, p, j):
            if custom_molar_term is not None:
                flow_basis = b.properties_out[t].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.molar:
                    return custom_molar_term(t, p, j)
                elif flow_basis == MaterialFlowBasis.mass:
                    try:
                        return (custom_molar_term(t, p, j) *
                                b.properties_out[t].mw[j])
                    except AttributeError:
                        raise PropertyNotSupportedError(
                                "{} property package does not support "
                                "molecular weight (mw), which is required for "
                                "using custom terms in material balances."
                                .format(self.name))
                else:
                    raise ConfigurationError(
                            "{} contained a custom_molar_term argument, but "
                            "the property package used an undefined basis "
                            "(MaterialFlowBasis.other). Custom terms can "
                            "only be used when the property package declares "
                            "a molar or mass flow basis.".format(self.name))
            else:
                return 0

        def user_term_mass(b, t, p, j):
            if custom_mass_term is not None:
                flow_basis = b.properties_out[t].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.mass:
                    return custom_mass_term(t, p, j)
                elif flow_basis == MaterialFlowBasis.molar:
                    try:
                        return (custom_mass_term(t, p, j) /
                                b.properties_out[t].mw[j])
                    except AttributeError:
                        raise PropertyNotSupportedError(
                                "{} property package does not support "
                                "molecular weight (mw), which is required for "
                                "using custom terms in material balances."
                                .format(self.name))
                else:
                    raise ConfigurationError(
                            "{} contained a custom_mass_term argument, but "
                            "the property package used an undefined basis "
                            "(MaterialFlowBasis.other). Custom terms can "
                            "only be used when the property package declares "
                            "a molar or mass flow basis.".format(self.name))
            else:
                return 0

        # Add component balances
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Material balances")
        def material_balances(b, t, p, j):
            if j in phase_component_list[p]:
                return accumulation_term(b, t, p, j) == (
                        b.properties_in[t].get_material_flow_terms(p, j) -
                        b.properties_out[t].get_material_flow_terms(p, j) +
                        kinetic_term(b, t, p, j) *
                        b._rxn_rate_conv(t, j, has_rate_reactions) +
                        equilibrium_term(b, t, p, j) +
                        phase_equilibrium_term(b, t, p, j) +
                        transfer_term(b, t, p, j) +
                        user_term_mol(b, t, p, j) + user_term_mass(b, t, p, j))
            else:
                return Constraint.Skip

        # TODO: Need to set material_holdup = 0 for non-present component-phase
        # pairs. Not ideal, but needed to close DoF. Is there a better way?

        # Material Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Material holdup calculations")
            def material_holdup_calculation(b, t, p, j):
                if j in phase_component_list[p]:
                    return b.material_holdup[t, p, j] == (
                          b.volume[t]*self.phase_fraction[t, p] *
                          b.properties_out[t].get_material_density_terms(p, j))
                else:
                    return b.material_holdup[t, p, j] == 0

        if has_rate_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.rate_reaction_extent = Var(
                    self.flowsheet().config.time,
                    self.config.reaction_package.rate_reaction_idx,
                    domain=Reals,
                    initialize=0.0,
                    doc="Extent of kinetic reactions[{}/{}]"
                        .format(units['holdup'], units['time']))

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Kinetic reaction stoichiometry constraint")
            def rate_reaction_stoichiometry_constraint(b, t, p, j):
                if j in phase_component_list[p]:
                    rparam = rblock[t]._params
                    return b.rate_reaction_generation[t, p, j] == (
                        sum(rparam.rate_reaction_stoichiometry[r, p, j] *
                            b.rate_reaction_extent[t, r]
                            for r in b.config.reaction_package.
                            rate_reaction_idx))
                else:
                    return Constraint.Skip

        if has_equilibrium_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.equilibrium_reaction_extent = Var(
                    self.flowsheet().config.time,
                    self.config.reaction_package.equilibrium_reaction_idx,
                    domain=Reals,
                    initialize=0.0,
                    doc="Extent of equilibrium reactions[{}/{}]"
                        .format(units['holdup'], units['time']))

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Equilibrium reaction stoichiometry")
            def equilibrium_reaction_stoichiometry_constraint(b, t, p, j):
                if j in phase_component_list[p]:
                    return b.equilibrium_reaction_generation[t, p, j] == (
                            sum(rblock[t]._params.
                                equilibrium_reaction_stoichiometry[r, p, j] *
                                b.equilibrium_reaction_extent[t, r]
                                for r in b.config.reaction_package.
                                equilibrium_reaction_idx))
                else:
                    return Constraint.Skip

        return self.material_balances

    def add_total_component_balances(self,
                                     has_rate_reactions=False,
                                     has_equilibrium_reactions=False,
                                     has_phase_equilibrium=False,
                                     has_mass_transfer=False,
                                     custom_molar_term=None,
                                     custom_mass_term=None):
        """
        This method constructs a set of 0D material balances indexed by time
        and component.

        Args:
            has_rate_reactions - whether default generation terms for rate
                    reactions should be included in material balances
            has_equilibrium_reactions - whether generation terms should for
                    chemical equilibrium reactions should be included in
                    material balances
            has_phase_equilibrium - whether generation terms should for phase
                    equilibrium behaviour should be included in material
                    balances
            has_mass_transfer - whether generic mass transfer terms should be
                    included in material balances
            custom_molar_term - a Pyomo Expression representing custom terms to
                    be included in material balances on a molar basis.
                    Expression must be indexed by time, phase list and
                    component list
            custom_mass_term - a Pyomo Expression representing custom terms to
                    be included in material balances on a mass basis.
                    Expression must be indexed by time, phase list and
                    component list

        Returns:
            Constraint object representing material balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        # Check that reaction block exists if required
        if has_rate_reactions or has_equilibrium_reactions:
            try:
                rblock = self.reactions
            except AttributeError:
                raise ConfigurationError(
                        "{} does not contain a Reaction Block, but material "
                        "balances have been set to contain reaction terms. "
                        "Please construct a reaction block before adding "
                        "balance equations.".format(self.name))

        if has_equilibrium_reactions:
            # Check that reaction block is set to calculate equilibrium
            for t in self.flowsheet().config.time:
                if self.reactions[t].config.has_equilibrium is False:
                    raise ConfigurationError(
                            "{} material balance was set to include "
                            "equilibrium reactions, however the associated "
                            "ReactionBlock was not set to include equilibrium "
                            "constraints (has_equilibrium_reactions=False). "
                            "Please correct your configuration arguments."
                            .format(self.name))

        if has_phase_equilibrium:
            # Check that state blocks are set to calculate equilibrium
            for t in self.flowsheet().config.time:
                if not self.properties_out[t].config.has_phase_equilibrium:
                    raise ConfigurationError(
                            "{} material balance was set to include phase "
                            "equilibrium, however the associated outlet "
                            "StateBlock was not set to include equilibrium "
                            "constraints (has_phase_equilibrium=False). Please"
                            " correct your configuration arguments."
                            .format(self.name))
                if not self.properties_in[t].config.has_phase_equilibrium:
                    raise ConfigurationError(
                            "{} material balance was set to include phase "
                            "equilibrium, however the associated inlet "
                            "StateBlock was not set to include equilibrium "
                            "constraints (has_phase_equilibrium=False). Please"
                            " correct your configuration arguments."
                            .format(self.name))

        # Get units from property package
        units = {}
        for u in ['length', 'holdup', 'amount', 'time']:
            try:
                units[u] = \
                   self.config.property_package.get_metadata().default_units[u]
            except KeyError:
                units[u] = '-'

        # Test for components that must exist prior to calling this method
        if has_holdup:
            if not hasattr(self, "volume"):
                raise ConfigurationError(
                        "{} control volume must have volume defined to have "
                        "holdup and/or rate reaction terms. Please call the "
                        "add_geometry method before adding balance equations."
                        .format(self.name))

        # Material holdup and accumulation
        if has_holdup:
            self.material_holdup = Var(
                    self.flowsheet().config.time,
                    self.config.property_package.phase_list,
                    self.config.property_package.component_list,
                    domain=Reals,
                    initialize=0.0,
                    doc="Material holdup in unit [{}]".format(units['holdup']))
        if dynamic:
            self.material_accumulation = DerivativeVar(
                    self.material_holdup,
                    wrt=self.flowsheet().config.time,
                    doc="Material accumulation in unit [{}/{}]"
                        .format(units['holdup'], units['time']))

        # Get phase component list(s)
        phase_component_list = self._get_phase_comp_list()

        # Create material balance terms as required
        # Kinetic reaction generation
        if has_rate_reactions:
            if not hasattr(self.config.reaction_package, "rate_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of rate "
                    "reactions (rate_reaction_idx), thus does not support "
                    "rate-based reactions.".format(self.name))
            self.rate_reaction_generation = Var(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated in "
                    "unit by kinetic reactions [{}/{}]"
                    .format(units['holdup'], units['time']))

        # Equilibrium reaction generation
        if has_equilibrium_reactions:
            if not hasattr(self.config.reaction_package,
                           "equilibrium_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of "
                    "equilibrium reactions (equilibrium_reaction_idx), thus "
                    "does not support equilibrium-based reactions."
                    .format(self.name))
            self.equilibrium_reaction_generation = Var(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated in unit "
                    "by equilibrium reactions [{}/{}]"
                    .format(units['holdup'], units['time']))

        # Material transfer term
        if has_mass_transfer:
            self.mass_transfer_term = Var(
                        self.flowsheet().config.time,
                        self.config.property_package.phase_list,
                        self.config.property_package.component_list,
                        domain=Reals,
                        initialize=0.0,
                        doc="Component material transfer into unit [{}/{}]"
                            .format(units['holdup'], units['time']))

        # Create rules to substitute material balance terms
        # Accumulation term
        def accumulation_term(b, t, p, j):
            return b.material_accumulation[t, p, j] if dynamic else 0

        def kinetic_term(b, t, p, j):
            return (b.rate_reaction_generation[t, p, j] if has_rate_reactions
                    else 0)

        def equilibrium_term(b, t, p, j):
            return (b.equilibrium_reaction_generation[t, p, j]
                    if has_equilibrium_reactions else 0)

        def transfer_term(b, t, p, j):
            return (b.mass_transfer_term[t, p, j] if has_mass_transfer else 0)

        def user_term_mol(b, t, j):
            if custom_molar_term is not None:
                flow_basis = b.properties_out[t].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.molar:
                    return custom_molar_term(t, j)
                elif flow_basis == MaterialFlowBasis.mass:
                    try:
                        return (custom_molar_term(t, j) *
                                b.properties_out[t].mw[j])
                    except AttributeError:
                        raise PropertyNotSupportedError(
                                "{} property package does not support "
                                "molecular weight (mw), which is required for "
                                "using custom terms in material balances."
                                .format(self.name))
                else:
                    raise ConfigurationError(
                            "{} contained a custom_molar_term argument, but "
                            "the property package used an undefined basis "
                            "(MaterialFlowBasis.other). Custom terms can "
                            "only be used when the property package declares "
                            "a molar or mass flow basis.".format(self.name))
            else:
                return 0

        def user_term_mass(b, t, j):
            if custom_mass_term is not None:
                flow_basis = b.properties_out[t].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.mass:
                    return custom_mass_term(t, j)
                elif flow_basis == MaterialFlowBasis.molar:
                    try:
                        return (custom_mass_term(t, j) /
                                b.properties_out[t].mw[j])
                    except AttributeError:
                        raise PropertyNotSupportedError(
                                "{} property package does not support "
                                "molecular weight (mw), which is required for "
                                "using custom terms in material balances."
                                .format(self.name))
                else:
                    raise ConfigurationError(
                            "{} contained a custom_mass_term argument, but "
                            "the property package used an undefined basis "
                            "(MaterialFlowBasis.other). Custom terms can "
                            "only be used when the property package declares "
                            "a molar or mass flow basis.".format(self.name))
            else:
                return 0

        # Add component balances
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.component_list,
                         doc="Material balances")
        def material_balances(b, t, j):
            cplist = []
            for p in self.config.property_package.phase_list:
                if j in phase_component_list[p]:
                    cplist.append(p)
            return (
                sum(accumulation_term(b, t, p, j) for p in cplist) ==
                sum(b.properties_in[t].get_material_flow_terms(p, j)
                    for p in cplist) -
                sum(b.properties_out[t].get_material_flow_terms(p, j)
                    for p in cplist) +
                sum(kinetic_term(b, t, p, j) for p in cplist) *
                b._rxn_rate_conv(t, j, has_rate_reactions) +
                sum(equilibrium_term(b, t, p, j) for p in cplist) +
                sum(transfer_term(b, t, p, j) for p in cplist) +
                user_term_mol(b, t, j) + user_term_mass(b, t, j))

        # TODO: Need to set material_holdup = 0 for non-present component-phase
        # pairs. Not ideal, but needed to close DoF. Is there a better way?

        # Material Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Material holdup calculations")
            def material_holdup_calculation(b, t, p, j):
                if j in phase_component_list[p]:
                    return b.material_holdup[t, p, j] == (
                          b.volume[t]*self.phase_fraction[t, p] *
                          b.properties_out[t].get_material_density_terms(p, j))
                else:
                    return b.material_holdup[t, p, j] == 0

        if has_rate_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.rate_reaction_extent = Var(
                    self.flowsheet().config.time,
                    self.config.reaction_package.rate_reaction_idx,
                    domain=Reals,
                    initialize=0.0,
                    doc="Extent of kinetic reactions[{}/{}]"
                        .format(units['holdup'], units['time']))

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Kinetic reaction stoichiometry constraint")
            def rate_reaction_stoichiometry_constraint(b, t, p, j):
                if j in phase_component_list[p]:
                    rparam = rblock[t]._params
                    return b.rate_reaction_generation[t, p, j] == (
                        sum(rparam.rate_reaction_stoichiometry[r, p, j] *
                            b.rate_reaction_extent[t, r]
                            for r in b.config.
                            reaction_package.rate_reaction_idx))
                else:
                    return Constraint.Skip

        if has_equilibrium_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.equilibrium_reaction_extent = Var(
                    self.flowsheet().config.time,
                    self.config.reaction_package.equilibrium_reaction_idx,
                    domain=Reals,
                    initialize=0.0,
                    doc="Extent of equilibrium reactions[{}/{}]"
                        .format(units['holdup'], units['time']))

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Equilibrium reaction stoichiometry")
            def equilibrium_reaction_stoichiometry_constraint(b, t, p, j):
                if j in phase_component_list[p]:
                    return b.equilibrium_reaction_generation[t, p, j] == (
                            sum(rblock[t]._params.
                                equilibrium_reaction_stoichiometry[r, p, j] *
                                b.equilibrium_reaction_extent[t, r]
                                for r in b.config.reaction_package.
                                equilibrium_reaction_idx))
                else:
                    return Constraint.Skip

        return self.material_balances

    def add_total_element_balances(self,
                                   has_rate_reactions=False,
                                   has_equilibrium_reactions=False,
                                   has_phase_equilibrium=False,
                                   has_mass_transfer=False,
                                   custom_elemental_term=None):
        """
        This method constructs a set of 0D element balances indexed by time.

        Args:
            has_rate_reactions - whether default generation terms for rate
                    reactions should be included in material balances
            has_equilibrium_reactions - whether generation terms should for
                    chemical equilibrium reactions should be included in
                    material balances
            has_phase_equilibrium - whether generation terms should for phase
                    equilibrium behaviour should be included in material
                    balances
            has_mass_transfer - whether generic mass transfer terms should be
                    included in material balances
            custom_elemental_term - a Pyomo Expression representing custom
                    terms to be included in material balances on a molar
                    elemental basis. Expression must be indexed by time and
                    element list

        Returns:
            Constraint object representing material balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        # Check that property package supports element balances
        if not hasattr(self.config.property_package, "element_list"):
            raise PropertyNotSupportedError(
                "{} property package provided does not contain a list of "
                "elements (element_list), and thus does not support "
                "elemental material balances. Please choose another type "
                "of material balance or a property package which supports "
                "elemental balances.")
        # Check for valid arguments to write total elemental balance
        if has_rate_reactions:
            raise ConfigurationError(
                "{} add_total_element_balances method provided with "
                "argument has_rate_reactions = True. Total element "
                "balances do not support rate based reactions, "
                "please correct your configuration arguments"
                .format(self.name))

        if has_equilibrium_reactions:
            raise ConfigurationError(
                "{} add_total_element_balances method provided with "
                "argument has_equilibrium_reactions = True. Total element "
                "balances do not support equilibrium based reactions, "
                "please correct your configuration arguments"
                .format(self.name))

        if has_phase_equilibrium:
            raise ConfigurationError(
                "{} add_total_element_balances method provided with "
                "argument has_phase_equilibrium = True. Total element "
                "balances do not support equilibrium based reactions, "
                "please correct your configuration arguments"
                .format(self.name))

        # Test for components that must exist prior to calling this method
        if has_holdup:
            if not hasattr(self, "volume"):
                raise ConfigurationError(
                        "{} control volume must have volume defined to have "
                        "holdup terms. Please call the "
                        "add_geometry method before adding balance equations."
                        .format(self.name))

        # Get units from property package
        units = {}
        for u in ['amount', 'time']:
            try:
                units[u] = \
                   self.config.property_package.get_metadata().default_units[u]
            except KeyError:
                units[u] = '-'

        # Add Material Balance terms
        if has_holdup:
            self.element_holdup = Var(
                    self.flowsheet().config.time,
                    self.config.property_package.element_list,
                    domain=Reals,
                    initialize=1.0,
                    doc="Elemental holdup in unit [{}]"
                        .format(units['amount']))

        if dynamic:
            self.element_accumulation = DerivativeVar(
                    self.element_holdup,
                    wrt=self.flowsheet().config.time,
                    doc="Elemental accumulation in unit [{}/{}]"
                        .format(units['amount'], units['time']))

        @self.Expression(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.element_list,
                         doc="Inlet elemental flow terms [{}/{}]"
                             .format(units['amount'], units['time']))
        def elemental_flow_in(b, t, p, e):
            return sum(b.properties_in[t].get_material_flow_terms(p, j) *
                       b.properties_out[t]._params.element_comp[j][e]
                       for j in b.config.property_package.component_list)

        @self.Expression(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.element_list,
                         doc="Outlet elemental flow terms [{}/{}]"
                             .format(units['amount'], units['time']))
        def elemental_flow_out(b, t, p, e):
            return sum(b.properties_out[t].get_material_flow_terms(p, j) *
                       b.properties_out[t]._params.element_comp[j][e]
                       for j in b.config.property_package.component_list)

        # Create material balance terms as needed
        if has_mass_transfer:
            self.elemental_mass_transfer_term = Var(
                            self.flowsheet().config.time,
                            self.config.property_package.element_list,
                            domain=Reals,
                            initialize=0.0,
                            doc="Element material transfer into unit [{}/{}]"
                            .format(units['amount'], units['time']))

        # Create rules to substitute material balance terms
        # Accumulation term
        def accumulation_term(b, t, e):
            return b.element_accumulation[t, e] if dynamic else 0

        # Mass transfer term
        def transfer_term(b, t, e):
            return (b.elemental_mass_transfer_term[t, e]
                    if has_mass_transfer else 0)

        # Custom term
        def user_term(t, e):
            if custom_elemental_term is not None:
                return custom_elemental_term(t, e)
            else:
                return 0

        # Element balances
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.element_list,
                         doc="Elemental material balances")
        def element_balances(b, t, e):
            return accumulation_term(b, t, e) == (
                        sum(b.elemental_flow_in[t, p, e]
                            for p in b.config.property_package.phase_list) -
                        sum(b.elemental_flow_out[t, p, e]
                            for p in b.config.property_package.phase_list) +
                        transfer_term(b, t, e) +
                        user_term(t, e))

        # Elemental Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.element_list,
                             doc="Elemental holdup calculation")
            def elemental_holdup_calculation(b, t, e):
                return b.element_holdup[t, e] == (
                    b.volume[t] *
                    sum(b.phase_fraction[t, p] *
                        b.properties_out[t].get_material_density_terms(p, j) *
                        b.properties_out[t]
                        ._params.element_comp[j][e]
                        for p in b.config.property_package.phase_list
                        for j in b.config.property_package.component_list))

        return self.element_balances

    def add_total_material_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_total_material_balances (yet)."
                .format(self.name))

    def add_total_enthalpy_balances(self,
                                    has_heat_of_reaction=False,
                                    has_heat_transfer=False,
                                    has_work_transfer=False,
                                    custom_term=None):
        """
        This method constructs a set of 0D enthalpy balances indexed by time
        and phase.

        Args:
            has_heat_of_reaction - whether terms for heat of reaction should
                    be included in enthalpy balance
            has_heat_transfer - whether terms for heat transfer should be
                    included in enthalpy balances
            has_work_transfer - whether terms for work transfer should be
                    included in enthalpy balances
            custom_term - a Pyomo Expression representing custom terms to
                    be included in enthalpy balances.
                    Expression must be indexed by time and phase list

        Returns:
            Constraint object representing enthalpy balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        # Test for components that must exist prior to calling this method
        if has_holdup:
            if not hasattr(self, "volume"):
                raise ConfigurationError(
                        "{} control volume must have volume defined to have "
                        "holdup terms. Please call the "
                        "add_geometry method before adding balance equations."
                        .format(self.name))
        if has_heat_of_reaction:
            if not (hasattr(self, "rate_reaction_extent") or
                    hasattr(self, "equilibrium_reaction_extent")):
                raise ConfigurationError(
                        "{} extent of reaction terms must exist in order to "
                        "have heat of reaction terms. Please ensure that "
                        "add_material_balance (or equivalent) is called before"
                        " adding energy balances.".format(self.name))

        # Get units from property package
        units = {}
        for u in ['energy', 'time']:
            try:
                units[u] = \
                   self.config.property_package.get_metadata().default_units[u]
            except KeyError:
                units[u] = '-'

        # Create variables
        if has_holdup:
            self.enthalpy_holdup = Var(
                        self.flowsheet().config.time,
                        self.config.property_package.phase_list,
                        domain=Reals,
                        initialize=1.0,
                        doc="Enthalpy holdup in unit [{}]"
                        .format(units['energy']))

        if dynamic is True:
            self.enthalpy_accumulation = DerivativeVar(
                        self.enthalpy_holdup,
                        wrt=self.flowsheet().config.time,
                        doc="Enthaly holdup in unit [{}/{}]"
                        .format(units['energy'], units['time']))

        # Create scaling factor
        self.scaling_factor_energy = Param(
                        default=1e-6,
                        mutable=True,
                        doc='Energy balance scaling parameter')

        # Create energy balance terms as needed
        # Heat transfer term
        if has_heat_transfer:
            self.heat = Var(self.flowsheet().config.time,
                            domain=Reals,
                            initialize=0.0,
                            doc="Heat transfered in unit [{}/{}]"
                                .format(units['energy'], units['time']))

        # Work transfer
        if has_work_transfer:
            self.work = Var(self.flowsheet().config.time,
                            domain=Reals,
                            initialize=0.0,
                            doc="Work transfered in unit [{}/{}]"
                                .format(units['energy'], units['time']))

        # Heat of Reaction
        if has_heat_of_reaction:
            @self.Expression(self.flowsheet().config.time,
                             doc="Heat of reaction term [{}/{}]"
                                 .format(units['energy'], units['time']))
            def heat_of_reaction(b, t):
                if hasattr(self, "rate_reaction_extent"):
                    rate_heat = -sum(b.rate_reaction_extent[t, r] *
                                     b.reactions[t].dh_rxn[r]
                                     for r in self.config.reaction_package.
                                     rate_reaction_idx)
                else:
                    rate_heat = 0

                if hasattr(self, "equilibrium_reaction_extent"):
                    equil_heat = -sum(
                            b.equilibrium_reaction_extent[t, e] *
                            b.reactions[t].dh_rxn[e]
                            for e in self.config.reaction_package.
                            equilibrium_reaction_idx)
                else:
                    equil_heat = 0

                return rate_heat + equil_heat

        # Create rules to substitute energy balance terms
        # Accumulation term
        def accumulation_term(b, t, p):
            return b.enthalpy_accumulation[t, p] if dynamic else 0

        def heat_term(b, t):
            return b.heat[t] if has_heat_transfer else 0

        def work_term(b, t):
            return b.work[t] if has_work_transfer else 0

        def rxn_heat_term(b, t):
            return b.heat_of_reaction[t] if has_heat_of_reaction else 0

        # Custom term
        def user_term(t):
            if custom_term is not None:
                return custom_term(t)
            else:
                return 0

        # Energy balance equation
        @self.Constraint(self.flowsheet().config.time, doc="Energy balances")
        def enthalpy_balances(b, t):
            return (sum(accumulation_term(b, t, p) for p in
                    b.config.property_package.phase_list) *
                    b.scaling_factor_energy) == (
                        sum(b.properties_in[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list) *
                        b.scaling_factor_energy -
                        sum(self.properties_out[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list) *
                        b.scaling_factor_energy +
                        heat_term(b, t)*b.scaling_factor_energy +
                        work_term(b, t)*b.scaling_factor_energy +
                        rxn_heat_term(b, t)*b.scaling_factor_energy +
                        user_term(t)*b.scaling_factor_energy)

        # Energy Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             doc="Enthalpy holdup constraint")
            def enthalpy_holdup_calculation(b, t, p):
                return b.enthalpy_holdup[t, p] == (
                            b.volume[t]*self.phase_fraction[t, p] *
                            b.properties_out[t].get_energy_density_terms(p))

        return self.enthalpy_balances

    def add_phase_enthalpy_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_phase_enthalpy_balances."
                .format(self.name))

    def add_phase_energy_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_phase_energy_balances."
                .format(self.name))

    def add_total_energy_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_total_energy_balances."
                .format(self.name))

    def add_total_pressure_balances(self,
                                    has_pressure_change=False,
                                    custom_term=None):
        """
        This method constructs a set of 0D pressure balances indexed by time.

        Args:
            has_pressure_change - whether terms for pressure change should be
                    included in enthalpy balances
            custom_term - a Pyomo Expression representing custom terms to
                    be included in pressure balances.
                    Expression must be indexed by time

        Returns:
            Constraint object representing pressure balances
        """
        # Get units from property package
        try:
            p_units = (self.config.property_package.get_metadata().
                       default_units['pressure'])
        except KeyError:
            p_units = '-'

        # Add Momentum Balance Variables as necessary
        if has_pressure_change:
            self.deltaP = Var(self.flowsheet().config.time,
                              domain=Reals,
                              initialize=0.0,
                              doc="Pressure difference across unit [{}]"
                                  .format(p_units))

        # Create rules to substitute energy balance terms
        # Pressure change term
        def deltaP_term(b, t):
            return b.deltaP[t] if has_pressure_change else 0

        # Custom term
        def user_term(t):
            if custom_term is not None:
                return custom_term(t)
            else:
                return 0

        # Create scaling factor
        self.scaling_factor_pressure = Param(
                    default=1e-4,
                    mutable=True,
                    doc='Momentum balance scaling parameter')

        # Momentum balance equation
        @self.Constraint(self.flowsheet().config.time, doc='Momentum balance')
        def pressure_balance(b, t):
            return 0 == (b.properties_in[t].pressure *
                         b.scaling_factor_pressure -
                         b.properties_out[t].pressure *
                         b.scaling_factor_pressure +
                         deltaP_term(b, t)*b.scaling_factor_pressure +
                         user_term(t)*b.scaling_factor_pressure)

        return self.pressure_balance

    def add_phase_pressure_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_phase_pressure_balances."
                .format(self.name))

    def add_phase_momentum_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_phase_momentum_balances."
                .format(self.name))

    def add_total_momentum_balances(self, *args, **kwargs):
        raise BalanceTypeNotSupportedError(
                "{} OD control volumes do not support "
                "add_total_momentum_balances."
                .format(self.name))

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().config.time:
            try:
                blk.properties_in[t].model_check()
            except AttributeError:
                _log.warning('{} ControlVolume inlet property block has no '
                             'model checks. To correct this, add a model_check'
                             ' method to the associated StateBlock class.'
                             .format(blk.name))
            try:
                blk.properties_out[t].model_check()
            except AttributeError:
                _log.warning('{} ControlVolume outlet property block has no '
                             'model checks. To correct this, add a '
                             'model_check method to the associated '
                             'StateBlock class.'.format(blk.name))

            try:
                blk.reactions[t].model_check()
            except AttributeError:
                _log.warning('{} ControlVolume outlet reaction block has no '
                             'model check. To correct this, add a '
                             'model_check method to the associated '
                             'ReactionBlock class.'.format(blk.name))

    def initialize(blk, state_args=None, outlvl=6, optarg=None,
                   solver='ipopt', hold_state=True):
        '''
        Initialization routine for 0D control volume (default solver ipopt)

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - True. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        '''
        # Get inlet state if not provided
        init_log = getInitLogger(blk.name, outlvl)
        if state_args is None:
            state_args = {}
            state_dict = (
                blk.properties_in[
                    blk.flowsheet().config.time.first()]
                .define_port_members())

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        # Initialize state blocks
        flags = blk.properties_in.initialize(outlvl=outlvl+1,
                                             optarg=optarg,
                                             solver=solver,
                                             hold_state=hold_state,
                                             state_args=state_args)

        blk.properties_out.initialize(outlvl=outlvl+1,
                                      optarg=optarg,
                                      solver=solver,
                                      hold_state=False,
                                      state_args=state_args)
        try:
            blk.reactions.initialize(outlvl=outlvl+1,
                                     optarg=optarg,
                                     solver=solver)
        except AttributeError:
            pass

        init_log.log(5, 'Initialization Complete')

        return flags

    def release_state(blk, flags, outlvl=6):
        '''
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        '''
        blk.properties_in.release_state(flags, outlvl=outlvl+1)

    def _add_phase_fractions(self):
        """
        This method constructs the phase_fraction variables for the control
        volume, and the associated constraint on the sum of phase_fractions
        == 1. For systems with only one phase, phase_fraction is created as a
        Pyomo Expression with a value of 1.

        Args:
            None

        Returns:
            None
        """
        if len(self.config.property_package.phase_list) > 1:
            self.phase_fraction = Var(
                            self.flowsheet().config.time,
                            self.config.property_package.phase_list,
                            initialize=1 / len(self.config.
                                               property_package.phase_list),
                            doc='Volume fraction of holdup by phase')

            @self.Constraint(self.flowsheet().config.time,
                             doc='Sum of phase fractions == 1')
            def sum_of_phase_fractions(self, t):
                return 1 == sum(self.phase_fraction[t, p]
                                for p in self.config.
                                property_package.phase_list)
        else:
            @self.Expression(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             doc='Volume fraction of holdup by phase')
            def phase_fraction(self, t, p):
                return 1

    def _rxn_rate_conv(b, t, j, has_rate_reactions):
        """
        Wrapper method for the _rxn_rate_conv method to hide the x argument
        required for 1D control volumes.
        """
        # Call the method in control_volume_base with x=None
        return super()._rxn_rate_conv(t, None, j, has_rate_reactions)

    def _get_performance_contents(self, time_point=0):
        """
        Collect all CV variables which are present to report.
        """
        var_dict = {}
        expr_dict = {}
        param_dict = {}

        time_only_vars = {"volume": "Volume",
                          "heat": "Heat Transfer",
                          "work": "Work Transfer",
                          "deltaP": "Pressure Change"}

        for v, n in time_only_vars.items():
            try:
                var_dict[n] = getattr(self, v)[time_point]
            except AttributeError:
                pass

        p_vars = {
            "phase_fraction": "Phase Fraction",
            "enthalpy_holdup": "Enthalpy Holdup",
            "enthalpy_accumulation": "Enthalpy Accumulation"}

        for v, n in p_vars.items():
            try:
                var_obj = getattr(self, v)
                for p in self.config.property_package.phase_list:
                    var_dict[f"{n} [{p}]"] = var_obj[time_point, p]
            except AttributeError:
                pass

        pc_vars = {
            "material_holdup": "Material Holdup",
            "material_accumulation": "Material Accumulation",
            "rate_reaction_generation": "Rate Reaction Generation",
            "equilibrium_reaction_generation":
                "Equilibrium Reaction Generation",
            "mass_transfer_term": "Mass Transfer Term"}

        for v, n in pc_vars.items():
            try:
                var_obj = getattr(self, v)
                for p in self.config.property_package.phase_list:
                    for j in self.config.property_package.component_list:
                        var_dict[f"{n} [{p}, {j}]"] = \
                            var_obj[time_point, p, j]
            except AttributeError:
                pass

        if hasattr(self, "rate_reaction_extent"):
            for r in self.config.reaction_package.rate_reaction_idx:
                var_dict[f"Rate Reaction Extent [{r}]"] = \
                    self.rate_reaction_extent[time_point, r]
        if hasattr(self, "equilibrium_reaction_extent"):
            for r in self.config.reaction_package.equilibrium_reaction_idx:
                var_dict[f"Equilibrium Reaction Extent [{r}]"] = \
                    self.equilibrium_reaction_extent[time_point, r]
        if hasattr(self, "phase_equilibrium_generation"):
            for r in self.config.property_package.phase_equilibrium_idx:
                var_dict[f"Phase Equilibrium Generation [{r}]"] = \
                    self.phase_equilibrium_generation[time_point, r]

        e_vars = {
            "element_holdup": "Elemental Holdup",
            "element_accumulation": "Elemental Accumulation",
            "elemental_mass_transfer_term": "Elemental Transfer Term"}

        for v, n in e_vars.items():
            try:
                var_obj = getattr(self, v)
                for e in self.config.property_package.element_list:
                    var_dict[f"{n} [{e}]"] = var_obj[time_point, e]
            except AttributeError:
                pass

        time_only_exprs = {"heat_of_reaction": "Heat of Reaction Term"}

        for e, n in time_only_exprs.items():
            try:
                expr_dict[n] = getattr(self, e)[time_point]
            except AttributeError:
                pass

        e_exprs = {"elemental_flow_in": "Element Flow In",
                   "elemental_flow_out": "Element Flow Out"}

        for o, n in e_exprs.items():
            try:
                expr_obj = getattr(self, o)
                for p in self.config.property_package.phase_list:
                    for e in self.config.property_package.element_list:
                        expr_dict[f"{n} [{p}, {e}]"] = \
                            expr_obj[time_point, p, e]
            except AttributeError:
                pass

        params = {"scaling_factor_energy": "Energy Scaling",
                  "scaling_factor_pressure": "Pressure Scaling"}

        for p, n in params.items():
            try:
                param_dict[n] = getattr(self, p)
            except AttributeError:
                pass

        return {"vars": var_dict,
                "exprs": expr_dict,
                "params": param_dict}

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe({"In": self.properties_in,
                                                  "Out": self.properties_out},
                                                 time_point=time_point)
        except AttributeError:
            return (f"Unit model {self.name} does not have the standard Port "
                    f"names (inet and outlet). Please contact the unit model "
                    f"developer to develop a unit specific stream table.")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
from collections import defaultdict
import json
import os

from idaes.core import UnitModelBlockData
from idaes.dmf.ui.link_position_mapping import link_position_mapping
from idaes.dmf.ui.icon_mapping import icon_mapping

# from pyomo.environ import Block
from pyomo.network.port import SimplePort
from pyomo.network.arc import SimpleArc


class FileBaseNameExistsError(Exception):
    pass


class FlowsheetSerializer:
    def __init__(self):
        self.unit_models = {}
        self.arcs = []
        self.ports = {}
        self.edges = defaultdict(list)
        self.orphaned_ports = {}

    def serialize(self, flowsheet, file_base_name, overwrite=False):
        """
        Serializes the flowsheet and saves it to a file that can be read by the
        idaes-model-vis  jupyter lab extension.

        :param flowsheet: The flowsheet to save. Usually fetched from the model.
        :param file_base_name: The file prefix to the .idaes.vis file produced.
        The file is created/saved
        in the directory that you ran from Jupyter Lab.
        :param overwrite: Boolean to overwrite an existing file_base_name.idaes.vis.
        If True, the existing file with the same file_base_name will be overwritten.
        This will cause you to lose
        any saved layout. 
        If False and there is an existing file with that file_base_name, you will get
        an error
        message stating that you cannot save a file to the file_base_name
        (and therefore overwriting the saved
        layout). If there is not an existing file with that file_base_name then it
        saves as normal.
        Defaults to False.
        :return: None

        Usage example:
            m = ConcreteModel()
            m.fs = FlowsheetBlock(...)
            ...
            serializer = FlowsheetSerializer()
            serializer.save(m.fs, "output_file")
        """
        vis_file_name = file_base_name + ".idaes.vis"
        if os.path.isfile(vis_file_name) and overwrite is False:
            msg = (
                f"{vis_file_name} already exists. If you wish to overwrite "
                f"this file call save() with overwrite=True. "
                "WARNING: If you overwrite the file, you will lose "
                "your saved layout."
            )
            raise FileBaseNameExistsError(msg)
        else:
            print(f"Creating {vis_file_name}")

        self.serialize_flowsheet(flowsheet)
        out_json = self._construct_output_json()

        with open(vis_file_name, "w") as out_file:
            json.dump(out_json, out_file)

    def _construct_output_json(self):
        out_json = {}
        out_json["cells"] = []
        x_pos = 50
        y_pos = 50

        for component, unit_attrs in self.unit_models.items():
            try:
                self.create_image_json(
                    out_json,
                    x_pos,
                    y_pos,
                    unit_attrs["name"],
                    icon_mapping[unit_attrs["type"]],
                    unit_attrs["name"],
                    unit_attrs["type"],
                )
            except KeyError:
                self.create_image_json(out_json, x_pos, y_pos, unit_attrs["name"], 
                                       "default", unit_attrs["name"], 
                                       unit_attrs["type"])

            x_pos += 50
            y_pos += 50

        id_counter = 0
        for source, dests in self.edges.items():
            umst = self.unit_models[source]["type"]  # alias
            for dest in dests:
                try:
                    if hasattr(source, "vap_outlet"):
                        # TODO Figure out how to denote different outlet types. Need to
                        #  deal with multiple input/output offsets
                        for arc in self.arcs:
                            if (
                                self.ports[arc.dest] == dest
                                and arc.source == source.vap_outlet
                            ):
                                source_anchor = link_position_mapping[umst][
                                    "top_outlet_anchor"
                                ]
                            elif (
                                self.ports[arc.dest] == dest
                                and arc.source == source.liq_outlet
                            ):
                                source_anchor = link_position_mapping[umst][
                                    "bottom_outlet_anchor"
                                ]

                    elif "top_outlet_anchor" in link_position_mapping[umst]:
                        source_anchor = \
                            link_position_mapping[umst]["top_outlet_anchor"]
                    else:
                        source_anchor = link_position_mapping[umst]["outlet_anchors"]
                        # TODO figure out offsets when mutiple things come
                        #  from/into the same side:
                        # source_anchor["args"]["dy"] = str(100/(len(dests) + 1)) + "%"

                except KeyError:
                    source_anchor = link_position_mapping["default"]["outlet_anchors"]
                    # TODO figure out offsets when mutiple things come from/into the 
                    # same side:
                    # source_anchor["args"]["dy"] = str(100/(len(dests) + 1)) + "%"
                try:
                    unit_type = self.unit_models[dest]["type"]
                    dest_anchor = \
                        link_position_mapping[unit_type]["inlet_anchors"]
                except KeyError:
                    dest_anchor = link_position_mapping["default"]["inlet_anchors"]

                self.create_link_json(
                    out_json, 
                    source_anchor, 
                    dest_anchor, 
                    source.getname(), 
                    dest.getname(), 
                    id_counter
                )
                id_counter += 1

        """
         TODO: We need a better way to define the inlets and outlets of the flowsheet 
         rather than just assuming the orphaned ports 
         are inlets and outlets. For now we are commenting out the orphaned port stuff.
        
        num_open_inlets = 0
        for orphan_port in self.orphaned_ports:
            unit_model_name = ""
            try:
                if num_open_inlets <= 0:
                    num_open_inlets = len(self.ports[orphan_port].create_inlet_list()) - 
                    len(self.edges[self.ports[orphan_port]])
            except AttributeError:
                num_open_inlets = 0
            if num_open_inlets > 0:
                icon_type = "feed"
            else:
                icon_type = "product"
            self.create_image_json(out_json, x_pos, y_pos, "inlet" + str(id_counter), 
            icon_mapping[icon_type], "", icon_type)
            x_pos += 50
            y_pos += 50
            try:
                dest_anchor = link_position_mapping[self.unit_models[
                self.ports[orphan_port]]
                ["type"]]["inlet_anchors"]
            except KeyError:
                dest_anchor = link_position_mapping["default"]["inlet_anchors"]
            if icon_type == "feed":
                source_anchor = link_position_mapping[icon_type]["outlet_anchors"]
                try:
                    dest_anchor = link_position_mapping[self.unit_models[
                    self.ports[orphan_port]]["type"]]["inlet_anchors"]
                except KeyError:
                    dest_anchor = link_position_mapping["default"]["inlet_anchors"]
                source_id = "inlet" + str(id_counter)
                dest_id = self.ports[orphan_port].getname()
            else:
                try:
                    source_anchor = link_position_mapping[self.unit_models[self.ports
                    [orphan_port]]["type"]]["inlet_anchors"]
                except KeyError:
                    source_anchor = link_position_mapping["default"]["inlet_anchors"]
                dest_anchor = link_position_mapping[icon_type]["outlet_anchors"]
                source_id = self.ports[orphan_port].getname()
                dest_id = "inlet" + str(id_counter)
            self.create_link_json(out_json, source_anchor, dest_anchor, source_id, 
            dest_id, id_counter)
            id_counter += 1
             num_open_inlets -= 1
        """

        return out_json

    def serialize_flowsheet(self, flowsheet):
        for component in flowsheet.component_objects(descend_into=False):
            # TODO try using component_objects(ctype=X)
            if isinstance(component, UnitModelBlockData):
                self.unit_models[component] = {
                    "name": component.getname(), 
                    "type": component._orig_module.split(".")[-1]
                }
                
                for subcomponent in component.component_objects(descend_into=True):
                    if isinstance(subcomponent, SimplePort):
                        self.ports[subcomponent] = component
                        
            elif isinstance(component, SimpleArc): 
                self.arcs.append(component)

        self.edges = defaultdict(list)
        # self.orphaned_ports = set(self.ports.keys())
        for arc in self.arcs:
            self.edges[self.ports[arc.source]].append(self.ports[arc.dest])
            # self.orphaned_ports.discard(arc.source)
            # self.orphaned_ports.discard(arc.dest)

    def create_image_json(self, out_json, x_pos, y_pos, id, image, label, title):
        entry = {}
        entry["type"] = "standard.Image"
        # for now, just tile the positions diagonally
        # TODO Make the default positioning better
        entry["position"] = {"x": x_pos, "y": y_pos}
        # TODO Set the width and height depending on the icon rather than default
        entry["size"] = {"width": 50, "height": 50}
        entry["angle"] = 0
        entry["id"] = id
        entry["z"] = (1,)
        entry["attrs"] = {
            "image": {"xlinkHref": image},
            "label": {"text": label},
            "root": {"title": title},
        }
        # out_json["model"]["cells"].append(entry)
        out_json["cells"].append(entry)

    def create_link_json(
        self, out_json, source_anchor, dest_anchor, source_id, dest_id, link_id
    ):
        entry = {
            "type": "standard.Link",
            "source": {"anchor": source_anchor, "id": source_id},
            "target": {"anchor": dest_anchor, "id": dest_id},
            "router": {"name": "orthogonal", "padding": 10},
            "connector": {"name": "jumpover", 
                          "attrs": {"line": {"stroke": "#6FB1E1"}}},
            "id": link_id,
            "z": 2,
            # "labels": [],
        }
        out_json["cells"].append(entry)

    def get_unit_models(self):
        return self.unit_models

    def get_ports(self):
        return self.ports

    def get_edges(self):
        return self.edges



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Set the alamo and gams paths here.
"""
# import collections as col


# Input-specific options and mechanisms
# Mechanisms used in CLC models are enumerated as strings are parsed later
# dlist old clcmlist
dlist = [
    "c1",
    "c2",
    "c3",
    "c4",
    "c5",
    "c6",
    "c7",
    "c8",
    "c9",
    "c10",
    "c11",
    "c12",
    "c13",
    "c14",
    "c15",
    "c16",
    "c17",
    "c18",
    "c19",
    "c20",
    "c21",
]
rspace = {}
rspace["mechanisms"] = list(["iT", "ma", "rev", "homcat", "clc", "ads"] + dlist)
rspace["clc"] = list(dlist)
rspace["clcnames"] = list(
    [
        "Random Nucleation",
        "Power law order 2/3",
        "Power law order 2",
        "Power law order 4",
        "Power law order 3",
        "Power law order 1",
        "Avrami order 2",
        "Avrami order 3",
        "Avrami order 0.5",
        "Avrami order 1.5",
        "Avrami order 4",
        "Prout Tompkins",
        "Jander 3d diff",
        "Anti Jander 3d diff",
        "Valensi",
        "Parabolic",
        "Ginstling-Brountstein 3d diff",
        "zhuralev lesohkin tempelman",
        "Grain Model",
    ]
)
rspace["clcforms"] = list(
    [
        "1-X",
        "(n=2/3) nX^(n-1/n)",
        "(n=2) nX^(n-1/n)",
        "(n=4) nX^(n-1/n)",
        "(n=3) nX^(n-1/n)",
        "(n=1) nX^(n-1/n)",
        "(n=2) n(1-X)(-ln(1-X))^(n-1/n)",
        "(n=3) n(1-X)(-ln(1-X))^(n-1/n)",
        "(n=0.5) n(1-X)(-ln(1-X))^(n-1/n)",
        "(n=1.5) n(1-X)(-ln(1-X))^(n-1/n)",
        "(n=4) n(1-X)(-ln(1-X))^(n-1/n)",
        "X(1-X)",
        "3(1-X)^(1/3)(1/(1+X)^((-1/3)-1))",
        "3/2(1-X)^(2/3)(1/(1+X)^((-1/3)-1))",
        "1/(-ln(1-X))",
        "1/(2X)",
        "(3/2)(1-X)^(4/3)((1-X)^((-1/3)-1))^-1",
        "(3/2)((1-X)^((-1/3)-1))^-1",
        "(1-X)^(2/3)",
    ]
)
# rspace['pclabels']=list(['T','v','x0','flow'])

# Initialize additional dictionaries, and initialize options that will be passed between subroutines
sharedata = {}
dlist = list(["profiles", "otherdata", "pwalamo"])
for item in dlist:
    sharedata[item] = {}
dlist = list(["profiles", "tlo", "tup", "etrack", "err", "nmin"])
for item in dlist:
    sharedata["pwalamo"][item] = {}
# Specify options pertaining to applicaiton of pw-alamo
sharedata["pwalamo"]["err"] = 0.99
sharedata["pwalamo"]["nmin"] = 3
sharedata["otherdata"]["eobj"] = 2
sharedata["gasconst"] = 0.008314  # J/mol K
sharedata["bounds"] = {}
sharedata["bounds"]["k"] = {}
sharedata["bounds"]["e"] = {}
sharedata["bounds"]["k"]["min"] = 0.0
sharedata["bounds"]["k"]["max"] = 500.0
sharedata["bounds"]["e"]["min"] = 0.0
sharedata["bounds"]["e"]["max"] = 200.0
sharedata["df"] = 1
sharedata["npc"] = 1
sharedata["ratemthd"] = "deriv"
# possible arguments for kwargs in ripemodel
sharedata["kwargsin"] = [
    "mechanisms",
    "mech",
    "mechs",
    "stoich",
    "stoichiometry",
    "stoichs",
    "minlp_path",
    "showpyomo",
    "alamo_path",
    "nlp_path",
    "keepfiles",
    "return_model",
    "Tref",
    "sigma",
    "hide_output",
    "ccon",
    "wls",
    "Tr",
    "deltaterm",
    "ascale",
    "zscale",
    "onemechper",
    "expand_output",
    "temp",
    "time",
    "tr",
    "tref",
    "temperature",
    "Temperature",
    "Temp",
]
# Set paths here
sharedata["minlp_path"] = "baron"
sharedata["nlp_path"] = "ipopt"
sharedata["alamo_path"] = "alamo"
sharedata["showpyomo"] = False
sharedata["keepfiles"] = False
sharedata["return_model"] = False
sharedata["hide_output"] = False
sharedata["wls"] = False
sharedata["adddepcons"] = False
sharedata["addfamcons"] = False
sharedata["maxmiptime"] = 1000.0
sharedata["deltaterm"] = 0
sharedata["zscale"] = False
sharedata["ascale"] = False
sharedata["onemechper"] = True
sharedata["expand_output"] = False

# independent variables considered by atermconstruct
sharedata["ivars"] = ["t", "T", "x0", "flow", "vol", "other"]


#
debug = {}
debug["savescratch"] = False
debug["true"] = {}
debug["bignum"] = 10 ** 20
debug["smallnum"] = 10 ** -10
debug["true"]["use"] = False
debug["true"]["fix"] = False



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains classes for reaction blocks and reaction parameter blocks.
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core.process_block import ProcessBlock
from idaes.core import ProcessBlockData
from idaes.core import property_meta
from idaes.core import MaterialFlowBasis
from idaes.core.util.exceptions import (BurntToast,
                                        PropertyNotSupportedError,
                                        PropertyPackageError)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_reaction_parameter_block,
                                    is_state_block)
from idaes.core.util.misc import add_object_reference

# Some more information about this module
__author__ = "Andrew Lee, John Eslick"

__all__ = ['ReactionBlockData',
           'ReactionBlock',
           'ReactionParameterBlock']


class ReactionParameterBlock(ProcessBlockData,
                             property_meta.HasPropertyClassMetadata):
    """
        This is the base class for reaction parameter blocks. These are blocks
        that contain a set of parameters associated with a specific reaction
        package, and are linked to by all instances of that reaction package.
    """
    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("property_package", ConfigValue(
            description="Reference to associated PropertyPackageParameter "
                        "object",
            domain=is_physical_parameter_block))
    CONFIG.declare("default_arguments", ConfigBlock(
            description="Default arguments to use with Property Package",
            implicit=True))

    def build(self):
        """
        General build method for ReactionParameterBlocks. Inheriting models
        should call super().build.

        Args:
            None

        Returns:
            None
        """
        super(ReactionParameterBlock, self).build()

        # TODO: Need way to tie reaction package to a specfic property package
        self._validate_property_parameter_units()
        self._validate_property_parameter_properties()

    def _validate_property_parameter_units(self):
        """
        Checks that the property parameter block associated with the
        reaction block uses the same set of default units.
        """
        r_units = self.get_metadata().default_units
        prop_units = self.config.property_package.get_metadata().default_units
        for u in r_units:
            try:
                if prop_units[u] != r_units[u]:
                    raise KeyError()
            except KeyError:
                raise PropertyPackageError(
                            '{} the property package associated with this '
                            'reaction package does not use the same set of '
                            'units of measurement ({}). Please choose a '
                            'property package which uses the same units.'
                            .format(self.name, u))

    def _validate_property_parameter_properties(self):
        """
        Checks that the property parameter block associated with the
        reaction block supports the necessary properties with correct units.
        """
        req_props = self.get_metadata().required_properties
        supp_props = self.config.property_package.get_metadata().properties

        for p in req_props:
            if p not in supp_props:
                raise PropertyPackageError(
                        '{} the property package associated with this '
                        'reaction package does not support the necessary '
                        'property, {}. Please choose a property package '
                        'which supports all required properties.'
                        .format(self.name, p))
            elif supp_props[p]['method'] is False:
                raise PropertyPackageError(
                        '{} the property package associated with this '
                        'reaction package does not support the necessary '
                        'property, {}. Please choose a property package '
                        'which supports all required properties.'
                        .format(self.name, p))

            # Check property units
            if req_props[p]['units'] != supp_props[p]['units']:
                raise PropertyPackageError(
                        '{} the units associated with property {} in this '
                        'reaction package ({}) do not match with the units '
                        'used in the assoicated property package ({}). Please '
                        'choose a property package which used the same '
                        'units for all properties.'
                        .format(self.name,
                                p,
                                req_props[p]['units'],
                                supp_props[p]['units']))


class ReactionBlockBase(ProcessBlock):
    """
        This is the base class for reaction block objects. These are used when
        constructing the SimpleBlock or IndexedBlock which will contain the
        PropertyData objects, and contains methods that can be applied to
        multiple ReactionBlockData objects simultaneously.
    """
    def initialize(self, *args):
        """
        This is a default initialization routine for ReactionBlocks to ensure
        that a routine is present. All ReactionBlockData classes should
        overload this method with one suited to the particular reaction package

        Args:
            None

        Returns:
            None
        """
        raise NotImplementedError('{} reaction package has not implemented the'
                                  ' initialize method. Please contact '
                                  'the reaction package developer'
                                  .format(self.name))

    def report(self, index=(0), true_state=False,
               dof=False, ostream=None, prefix=""):
        raise NotImplementedError(
                """The current Reaction Package has not implemented a report
                method. Please contact the package developer about this.""")


class ReactionBlockDataBase(ProcessBlockData):
    """
        This is the base class for reaction block data objects. These are
        blocks that contain the Pyomo components associated with calculating a
        set of reacion properties for a given material.
    """
    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("parameters", ConfigValue(
            domain=is_reaction_parameter_block,
            description="""A reference to an instance of the Reaction Parameter
Block associated with this property package."""))
    CONFIG.declare("state_block", ConfigValue(
            domain=is_state_block,
            description="""A reference to an instance of a StateBlock with
which this reaction block should be associated."""))
    CONFIG.declare("has_equilibrium", ConfigValue(
            default=True,
            domain=In([True, False]),
            description="Equilibrium constraint flag",
            doc="""Flag indicating whether equilibrium constraints
should be constructed in this reaction block,
**default** - True.
**Valid values:** {
**True** - ReactionBlock should enforce equilibrium constraints,
**False** - ReactionBlock should not enforce equilibrium constraints.}"""))

    def build(self):
        """
        General build method for PropertyBlockDatas. Inheriting models should
        call super().build.

        Args:
            None

        Returns:
            None
        """
        super(ReactionBlockDataBase, self).build()
        add_object_reference(self, "_params", self.config.parameters)

        self._validate_state_block()

    def _validate_state_block(self):
        """
        Method to validate that the associated state block matches with the
        PropertyParameterBlock assoicated with the ReactionParameterBlock.
        """
        # Add a reference to the corresponding state block data for later use
        add_object_reference(self,
                             "state_ref",
                             self.config.state_block[self.index()])

        # Validate that property package of state matches that of reaction pack
        if (self.config.parameters.config.property_package !=
                self.state_ref.config.parameters):
            raise PropertyPackageError(
                            '{} the StateBlock associated with this '
                            'ReactionBlock does not match with the '
                            'PropertyParamterBlock associated with the '
                            'ReactionParameterBlock. The modelling framework '
                            'does not support mixed associations of property '
                            'and reaction packages.'
                            .format(self.name))

    def get_reaction_rate_basis(self):
        """
        Method which returns an Enum indicating the basis of the reaction rate
        term.
        """
        return MaterialFlowBasis.other

    def __getattr__(self, attr):
        """
        This method is used to avoid generating unnecessary property
        calculations in reaction blocks. __getattr__ is called whenever a
        property is called for, and if a propery does not exist, it looks for
        a method to create the required property, and any associated
        components.

        Create a property calculation if needed. Return an attrbute error if
        attr == 'domain' or starts with a _ . The error for _ prevents a
        recursion error if trying to get a function to create a property and
        that function doesn't exist.  Pyomo also ocasionally looks for things
        that start with _ and may not exist.  Pyomo also looks for the domain
        attribute, and it may not exist.
        This works by creating a property calculation by calling the "_"+attr
        function.

        A list of __getattr__ calls is maintained in self.__getattrcalls to
        check for recursive loops which maybe useful for debugging. This list
        is cleared after __getattr__ completes successfully.

        Args:
            attr: an attribute to create and return. Should be a property
                  component.
        """

        def clear_call_list(self, attr):
            """Local method for cleaning up call list when a call is handled.

                Args:
                    attr: attribute currently being handled
            """
            if self.__getattrcalls[-1] == attr:
                if len(self.__getattrcalls) <= 1:
                    del self.__getattrcalls
                else:
                    del self.__getattrcalls[-1]
            else:
                raise PropertyPackageError(
                        "{} Trying to remove call {} from __getattr__"
                        " call list, however this is not the most "
                        "recent call in the list ({}). This indicates"
                        " a bug in the __getattr__ calls. Please "
                        "contact the IDAES developers with this bug."
                        .format(self.name, attr, self.__getattrcalls[-1]))

        # Check that attr is not something we shouldn't touch
        if attr == "domain" or attr.startswith("_"):
            # Don't interfere with anything by getting attributes that are
            # none of my business
            raise PropertyPackageError(
                    '{} {} does not exist, but is a protected '
                    'attribute. Check the naming of your '
                    'components to avoid any reserved names'
                    .format(self.name, attr))

        if attr == "config":
            try:
                self._get_config_args()
                return self.config
            except:
                raise BurntToast("{} getattr method was triggered by a call "
                                 "to the config block, but _get_config_args "
                                 "failed. This should never happen.")

        # Check for recursive calls
        try:
            # Check to see if attr already appears in call list
            if attr in self.__getattrcalls:
                # If it does, indicates a recursive loop.
                if attr == self.__getattrcalls[-1]:
                    # attr method is calling itself
                    self.__getattrcalls.append(attr)
                    raise PropertyPackageError(
                                    '{} _{} made a recursive call to '
                                    'itself, indicating a potential '
                                    'recursive loop. This is generally '
                                    'caused by the {} method failing to '
                                    'create the {} component.'
                                    .format(self.name, attr, attr, attr))
                else:
                    self.__getattrcalls.append(attr)
                    raise PropertyPackageError(
                                    '{} a potential recursive loop has been '
                                    'detected whilst trying to construct {}. '
                                    'A method was called, but resulted in a '
                                    'subsequent call to itself, indicating a '
                                    'recursive loop. This may be caused by a '
                                    'method trying to access a component out '
                                    'of order for some reason (e.g. it is '
                                    'declared later in the same method). See '
                                    'the __getattrcalls object for a list of '
                                    'components called in the __getattr__ '
                                    'sequence.'
                                    .format(self.name, attr))
            # If not, add call to list
            self.__getattrcalls.append(attr)
        except AttributeError:
            # A list of calls if one does not exist, so create one
            self.__getattrcalls = [attr]

        # Get property information from get_supported_properties
        try:
            m = self.config.parameters.get_metadata().properties

            if m is None:
                raise PropertyPackageError(
                        '{} reaction package get_supported_properties'
                        ' method returned None when trying to create '
                        '{}. Please contact the developer of the '
                        'property package'.format(self.name, attr))
        except KeyError:
            # If attr not in get_supported_properties, assume package does not
            # support property
            clear_call_list(self, attr)
            raise PropertyNotSupportedError(
                    '{} {} is not supported by reaction package (property is '
                    'not listed in get_supported_properties).'
                    .format(self.name, attr, attr))

        # Get method name from get_supported_properties
        try:
            if m[attr]['method'] is None:
                # If method is none, property should be constructed
                # by property package, so raise PropertyPackageError
                clear_call_list(self, attr)
                raise PropertyPackageError(
                        '{} {} should be constructed automatically '
                        'by reaction package, but is not present. '
                        'This can be caused by methods being called '
                        'out of order.'.format(self.name, attr))
            elif m[attr]['method'] is False:
                # If method is False, package does not support property
                # Raise NotImplementedError
                clear_call_list(self, attr)
                raise PropertyNotSupportedError(
                        '{} {} is not supported by reaction package '
                        '(property method is listed as False in '
                        'get_supported_properties).'
                        .format(self.name, attr))
            elif isinstance(m[attr]['method'], str):
                # Try to get method name in from PropertyBlock object
                try:
                    f = getattr(self, m[attr]['method'])
                except AttributeError:
                    # If fails, method does not exist
                    clear_call_list(self, attr)
                    raise PropertyPackageError(
                            '{} {} get_supported_properties method '
                            'returned a name that does not correspond'
                            ' to any method in the reaction package. '
                            'Please contact the developer of the '
                            'reaction package.'.format(self.name, attr))
            else:
                # Otherwise method name is invalid
                clear_call_list(self, attr)
                raise PropertyPackageError(
                             '{} {} get_supported_properties method '
                             'returned invalid value for method name. '
                             'Please contact the developer of the '
                             'reaction package.'
                             .format(self.name, attr))
        except KeyError:
            # No method key - raise Exception
            # Need to use an AttributeError so Pyomo.DAE will handle this
            clear_call_list(self, attr)
            raise PropertyNotSupportedError(
                    '{} get_supported_properties method '
                    'does not contain a method for {}. '
                    'Please select a package which supports '
                    'the necessary properties for your process.'
                    .format(self.name, attr))

        # Call attribute if it is callable
        # If this fails, it should return a meaningful error.
        if callable(f):
            try:
                f()
            except Exception:
                # Clear call list and reraise error
                clear_call_list(self, attr)
                raise
        else:
            # If f is not callable, inform the user and clear call list
            clear_call_list(self, attr)
            raise PropertyPackageError(
                    '{} tried calling attribute {} in order to create '
                    'component {}. However the method is not callable.'
                    .format(self.name, f, attr))

        # Clear call list, and return
        comp = getattr(self, attr)
        clear_call_list(self, attr)
        return comp



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
from idaes.surrogate import ripe
import numpy as np
# import itertools as it
import sys


def ripemodel(data, **kwargs):
    # import pyomo.environ

    #    from ripe import mechs, kinforms, setshared
    # import os

    sharedata, debug = ripe.sharedata, ripe.debug
    sharedata, kwargs = ripe.checkoptions(kwargs, sharedata, debug)

    # input to ripe specifies knietic mechanisms for reaction stoichiometries
    # Get mechs formats or creates these kinetic mechanisms
    stoich, mechanisms, fixarray, ncons, mechlist = ripe.mechs.getmechs(kwargs)

    # Find dependent stoichiometries

    if False:
        #    if np.linalg.matrix_rank(np.array(stoich)) < len(stoich):
        sharedata["adddepcons"] = True
        dep_small, dep_large, ncon, fam_dep, nfam = ripe.bounds.stoich_cons(stoich)
        if len(ncon) > 0:
            sharedata["adddepcons"] = True
        if len(nfam) > 0:
            sharedata["addfamcons"] = True
    else:
        dep_small, dep_large, ncon, fam_dep, nfam = [False, False, False, False, False]
        sharedata["adddepcons"] = False
        sharedata["addfamcons"] = False
    sharedata["sdep"] = [dep_small, dep_large, ncon, fam_dep, nfam]

    # Construction of the activity term is handled here
    # i.e. - rate = k exp(-E/RT) * Aterm
    aterm, fdata, pc, data, scales = ripe.atermconstruct.makeaterm(
        data, stoich, mechanisms, kwargs, ncons, mechlist, fixarray, sharedata
    )
    # remove stoichiometric-dependent flag from mechs
    mechanisms = mechanisms[0]

    # Aterm is a 4-dimensional array where
    # n = number of process conditions (t included for dynamic data)
    # ns = number of species
    # nm = number of mechanisms
    # nh = number of stoichiometries
    n, ns, nm, nh = np.shape(aterm)

    # Need to calculate target values
    targets, s_targets, savetargets, sharedata = ripe.targets.gentargets(
        data, kwargs, fdata, pc, sharedata
    )

    # ptype determines arrhenious vs linear kinetic models

    if "T" in kwargs.keys():
        ptype = "arr"
    elif ns > 1:
        ptype = "iT"
    else:
        ptype = "simple"

    # Initialize sigma then calculate variance of residuals
    if "sigma" in kwargs.keys():
        if isinstance(kwargs["sigma"], type(0.0)) or isinstance(kwargs["sigma"], type(0)):
            kwargs["sigma"] = np.array(kwargs["sigma"])
        sigma = np.squeeze(kwargs["sigma"])
        dshape = np.shape(sigma)
        if len(dshape) == 1:
            # First determine if the sigma specified is a single value or vector
            if dshape[0] == 1:
                sigma = np.multiply(np.ones([n, ns]), sigma[0])
                sharedata["wls"] = False
            elif dshape[0] == ns:
                sigma = np.multiply(np.ones([n, ns]), sigma[:])
                sharedata["wls"] = False
            else:
                sys.exit("Invalid shape specified for input argument sigma")
        elif len(dshape) == 2:
            sigma = np.reshape(sigma, (n, ns))
            sharedata["wls"] = True

        # Replace specified zero sigma with small number
        for i in range(dshape[0]):
            for j in range(dshape[1]):
                if sigma[i, j] < 0.0:
                    sys.exit("Negative varaince defined by user")
                elif sigma[i, j] == 0.0:
                    sigma[i, j] = 10 ** -8
        # Initial call of ripeomo to get refined bounds
        sharedata = ripe.bounds.get_bounds(
            [
                [aterm, targets, sigma, sharedata["bounds"]],
                ptype,
                fixarray,
                ncons,
                0,
                pc,
                sharedata,
            ]
        )
        # Need to claculate refined variable bounds here
    else:
        sigma = [[1.0] * int(ns)] * int(n)
        ccon = np.min([ns, ncons, n, nm, nh])
        sys.stdout.write(
            "Calcuating residual variance with data provided with cc = "
            + str(ccon)
            + "\n"
        )
        sigresults = ripe.genpyomo.ripeomo(
            [aterm, targets, sigma, sharedata["bounds"]],
            ptype,
            fixarray,
            ccon,
            0,
            pc,
            sharedata,
        )
        sigma = [sigresults["sigma"] * int(ns)] * int(n)
        if sharedata["wls"]:
            sigma = sigresults["wlsmat"]
    sigma = np.array(sigma)
    dshape = np.shape(sigma)

    for i in range(dshape[0]):
        for j in range(dshape[1]):
            if sigma[i, j] < 10 ** -8:
                sigma[i, j] = 10 ** -8

    # w_norm = sum(sum(sigma))
    savesig = sigma
    if sharedata["zscale"]:
        for i in range(n):
            for j in range(ns):
                sigma[i, j] = sigma[i, j] / (s_targets ** 2)
            # if sigma[i,j] < debug['smallnum'] or savesig

    if "ccon" in kwargs.keys():
        ccon = int(kwargs["ccon"][0][0])
        sys.stdout.write(
            "Solving RIPE model with cardinality constraint = " + str(ccon) + "\n"
        )
        results = ripe.genpyomo.ripeomo(
            [aterm, targets, sigma, sharedata["bounds"]],
            ptype,
            fixarray,
            ccon,
            1,
            pc,
            sharedata,
        )
    else:
        # ccon is not specified, use BIC to size model
        if ptype == "arr":
            ccon_list = range(1, 1 + np.min([nh, ns, int(np.floor((n - 1) / 2))]))
        else:
            ccon_list = range(1, 1 + np.min([nh, ns, int(n - 1)]))
        bic = []
        # Overwriting results variable here
        d_results = dict.fromkeys([0] + list(ccon_list)) 
        sys.stdout.write(
            "   ---- Calculating null values for model selection ----    \n"
        )
        # Getting sigma estimates from ~unbiased estimates
        d_results[0] = ripe.genpyomo.ripeomo(
            [aterm, targets, sigma, sharedata["bounds"]],
            ptype,
            fixarray,
            0,
            1,
            pc,
            sharedata,
        )
        bic.append(d_results[0]["OBJ"])
        sys.stdout.write(" - Null model BIC = " + str(bic[-1]) + "\n")
        for ccon in ccon_list:
            sys.stdout.write(
                " - Solving RIPE model with cardinality constraint = "
                + str(ccon)
                + " - \n"
            )
            d_results[ccon] = ripe.genpyomo.ripeomo(
                [aterm, targets, sigma, sharedata["bounds"]],
                ptype,
                fixarray,
                ccon,
                1,
                pc,
                sharedata,
            )
            bic.append(float(d_results[ccon]["OBJ"]) + np.log(n * ns) * ccon)
            sys.stdout.write(
                " - " + str(ccon) + "-term model BIC = " + str(bic[-1]) + "\n"
            )
            if bic[-1] > bic[-2]:
                ccon = ccon - 1
                break

        results = d_results[ccon]

    if "sigma" not in kwargs.keys():
        sigma = np.array(results["wlsmat"])
    else:
        sigma = savesig
    kres, eres, stoichres, mechres, r2 = ripe.confinv(
        aterm, targets, results, ccon, ptype, pc, sharedata, sigma
    )  # ,scales,s_targets)
    #    kres, eres, stoichres, mechres,r2 = ripe.confinv(scales['aterm_u'],savetargets,results,ccon,ptype,pc,sharedata,sigma,scales,s_targets)
    for i in range(len(mechres)):
        in1 = int(stoichres[i]) - 1
        in2 = int(mechres[i]) - 1
        stoichres[i] = stoich[in1]
        mechres[i] = mechlist[in2]
        if sharedata["ascale"]:
            kres[i][0] = kres[i][0] / float(scales["nfactor"][in2, in1])
            kres[i][1] = kres[i][1] / float(scales["nfactor"][in2, in1])
        if sharedata["zscale"]:
            kres[i][0] = s_targets * kres[i][0]
            kres[i][1] = s_targets * kres[i][1]

    s_kres, s_eres, ci_k, ci_e = print_results(
        kres, eres, stoichres, mechres, r2, sharedata
    )

    main_return = {}
    # Additional information is included if expand_outputs is true
    if sharedata["expand_output"]:
        main_return["wls"] = results["wlsmat"]
        if "sigma" in kwargs.keys():
            main_return["sigma"] = sigma
        else:
            main_return["sigma"] = results["sigma"]

    main_return["k"] = s_kres
    main_return["conf_inv"] = ci_k
    main_return["stoichiometry"] = stoichres
    main_return["mechanisms"] = mechres
    if eres != []:
        main_return["E"] = s_eres
        main_return["conf_inv"].append([ci_e])
    return main_return


def ripewrite(sd, string):
    import sys

    with open("riperesults.txt", "a") as a:
        a.write(string + "\n")
    if not sd["hide_output"]:
        sys.stdout.write(string + "\n")


def print_results(kres, eres, stoichres, mechres, r2, sd):
    import sys

    cc = len(mechres)

    s_kres = []
    s_eres = []
    ci_k = []
    ci_e = []
    tstring = (
        "  ----  RIPE selected "
        + str(cc)
        + " models in the optimal reaction network  ----  \n"
    )
    sys.stdout.write(tstring)
    with open("riperesults.txt", "w") as a:
        a.write(tstring)
    ripewrite(sd, " -+-  Overall R2 of selected model : " + str(r2))
    for i in range(cc):
        ripewrite(sd, " - Stochiometry selected for reaction " + str(i + 1))
        ripewrite(sd, "         " + str(stoichres[i]))
        ripewrite(sd, "   Mechanism selected for reaction " + str(i + 1))
        ripewrite(sd, "         " + str(mechres[i]))
        ripewrite(sd, "   Kinetic rate constant & 95% confidence interval")
        ripewrite(sd, "         " + str(kres[i][0]) + " +/- " + str(kres[i][1]))
        ci_k.append(kres[i][1])
        s_kres.append(kres[i][0])
        if eres != []:
            ripewrite(sd, "   Activation energy & 95% confidence interval")
            ripewrite(sd, "        " + str(eres[i][0]) + " +/- " + str(eres[i][1]))
            s_eres.append(eres[i][1])
            ci_e.append(eres[i][1])
        ripewrite(sd, "")

    return [s_kres, s_eres, ci_k, ci_e]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# These subroutines ensure compatibility across linux and windows operating systems


import os
import platform
import subprocess


def deletefile(*fname):
    "Deletes files"
    tos = platform.platform()
    currentDirectory = os.getcwd()
    if 'Windows' in tos:
        for name in fname:
            os.system("del %s/" % currentDirectory + name)
    else:
        for name in fname:
            os.system("rm %s/" % currentDirectory + name)


def movefile(*fname):
    "Moves files"
    tos = platform.platform()
    if 'Windows' in tos:
        for name in fname:
            os.system("move " + name)
    else:
        for name in fname:
            os.system("mv " + name)


def copyfile(outf, inf):
    "Copies files"
    tos = platform.platform()
    if 'Windows' in tos:
        os.system("copy " + inf + ' ' + outf)
    else:
        os.system("cp " + inf + ' ' + outf)


def catfile(outf, *fname):
    "Concatenates files"
    tos = platform.platform()
    if 'Windows' in tos:
        ostr = ''
        for name in fname:
            ostr += "echo " + name + ' >> ' + outf + ' & '
        ostr = ostr[:-3]
        os.system(ostr)
    else:
        ostr = 'cat '
        for name in fname:
            ostr += name + ' '
        ostr += '> ' + outf
        os.system(ostr)


def has_alamo():
    """Checks for ALAMO
    """
    try:
        s = subprocess.check_output(["alamo"])
        if b"Licensing error" in s: 
            _alamo_ok = False
        else:
            _alamo_ok = True
    except (subprocess.CalledProcessError, FileNotFoundError):
        _alamo_ok = False

    return _alamo_ok



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Heat Exchanger Models.
"""

__author__ = "John Eslick"

from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Param,
    log,
    Expression,
    Constraint,
    Reference,
    PositiveReals,
    SolverFactory,
    ExternalFunction,
    exp,
    log10,
    Block,
    Reference,
)

from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.opt import TerminationCondition

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    MaterialBalanceType,
    UnitModelBlockData,
    useDefault,
)

from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition
from idaes.functions import functions_lib
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.unit_models.heater import (
    _make_heater_config_block,
    _make_heater_control_volume,
)

import idaes.core.util.unit_costing as costing
from idaes.core.util.misc import add_object_reference

_log = getIdaesLogger(__name__)


class HeatExchangerFlowPattern(Enum):
    countercurrent = 1
    cocurrent = 2
    crossflow = 3


def _make_heat_exchanger_config(config):
    """
    Declare configuration options for HeatExchangerData block.
    """
    config.declare(
        "hot_side_name",
        ConfigValue(
            default="shell",
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "cold_side_name",
        ConfigValue(
            default="tube",
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "hot_side_config",
        ConfigBlock(
            implicit=True,
            description="Config block for hot side",
            doc="""A config block used to construct the hot side control volume.
This config can be given by the hot side name instead of hot_side_config.""",
        ),
    )
    config.declare(
        "cold_side_config",
        ConfigBlock(
            implicit=True,
            description="Config block for cold side",
            doc="""A config block used to construct the cold side control volume.
This config can be given by the cold side name instead of cold_side_config.""",
        ),
    )
    _make_heater_config_block(config.hot_side_config)
    _make_heater_config_block(config.cold_side_config)
    config.declare(
        "delta_temperature_callback",
        ConfigValue(
            default=delta_temperature_lmtd_callback,
            description="Callback for for temperature difference calculations",
        ),
    )
    config.declare(
        "flow_pattern",
        ConfigValue(
            default=HeatExchangerFlowPattern.countercurrent,
            domain=In(HeatExchangerFlowPattern),
            description="Heat exchanger flow pattern",
            doc="""Heat exchanger flow pattern,
**default** - HeatExchangerFlowPattern.countercurrent.
**Valid values:** {
**HeatExchangerFlowPattern.countercurrent** - countercurrent flow,
**HeatExchangerFlowPattern.cocurrent** - cocurrent flow,
**HeatExchangerFlowPattern.crossflow** - cross flow, factor times
countercurrent temperature difference.}""",
        ),
    )


def delta_temperature_lmtd_callback(b):
    """
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD).  It can be supplied to "delta_temperature_callback"
    HeatExchanger configuration option.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().config.time)
    def delta_temperature(b, t):
        return (dT1[t] - dT2[t]) / log(dT1[t] / dT2[t])


def delta_temperature_amtd_callback(b):
    """
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using arithmetic-mean
    temperature difference (AMTD).  It can be supplied to
    "delta_temperature_callback" HeatExchanger configuration option.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().config.time)
    def delta_temperature(b, t):
        return (dT1[t] + dT2[t]) * 0.5


def delta_temperature_underwood_callback(b):
    """
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD) approximation given by Underwood (1970).  It can be
    supplied to "delta_temperature_callback" HeatExchanger configuration option.
    This uses a cube root function that works with negative numbers returning
    the real negative root. This should always evaluate successfully.
    """
    # external function that ruturns the real root, for the cuberoot of negitive
    # numbers, so it will return without error for positive and negitive dT.
    b.cbrt = ExternalFunction(library=functions_lib(), function="cbrt")
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().config.time)
    def delta_temperature(b, t):
        return ((b.cbrt(dT1[t]) + b.cbrt(dT2[t])) / 2.0) ** 3


@declare_process_block_class("HeatExchanger", doc="Simple 0D heat exchanger model.")
class HeatExchangerData(UnitModelBlockData):
    """
    Simple 0D heat exchange unit.
    Unit model to transfer heat from one material to another.
    """

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    _make_heat_exchanger_config(CONFIG)

    def set_scaling_factor_energy(self, f):
        """
        This function sets scaling_factor_energy for both side_1 and side_2.
        This factor multiplies the energy balance and heat transfer equations
        in the heat exchnager.  The value of this factor should be about
        1/(expected heat duty).

        Args:
            f: Energy balance scaling factor
        """
        self.side_1.scaling_factor_energy.value = f
        self.side_2.scaling_factor_energy.value = f

    def _process_config(self):
        """Check for configuration errors and alternate config option names.
        """
        config = self.config

        if config.hot_side_name == config.cold_side_name:
            raise NameError(
                "Heatexchanger hot and cold side cannot have the same name '{}'."
                " Be sure to set both the hot_side_name and cold_side_name.".format(
                    config.hot_side_name
                )
            )

        for o in config:
            if not (
                o in self.CONFIG or o in [config.hot_side_name, config.cold_side_name]
            ):
                raise KeyError("Heatexchanger config option {} not defined".format(o))

        if config.hot_side_name in config:
            config.hot_side_config.set_value(config[config.hot_side_name])
            # Allow access to hot_side_config under the hot_side_name, backward
            # compatible with the tube and shell notation
            setattr(config, config.hot_side_name, config.hot_side_config)
        if config.cold_side_name in config:
            config.cold_side_config.set_value(config[config.cold_side_name])
            # Allow access to hot_side_config under the cold_side_name, backward
            # compatible with the tube and shell notation
            setattr(config, config.cold_side_name, config.cold_side_config)

    def build(self):
        """
        Building model

        Args:
            None
        Returns:
            None
        """
        ########################################################################
        #  Call UnitModel.build to setup dynamics and configure                #
        ########################################################################
        super().build()
        self._process_config()
        config = self.config

        ########################################################################
        # Add variables                                                        #
        ########################################################################
        u = self.overall_heat_transfer_coefficient = Var(
            self.flowsheet().config.time,
            domain=PositiveReals,
            initialize=100.0,
            doc="Overall heat transfer coefficient",
        )
        a = self.area = Var(
            domain=PositiveReals, initialize=1000.0, doc="Heat exchange area"
        )
        self.delta_temperature_in = Var(
            self.flowsheet().config.time,
            initialize=10.0,
            doc="Temperature difference at the hot inlet end",
        )
        self.delta_temperature_out = Var(
            self.flowsheet().config.time,
            initialize=10.0,
            doc="Temperature difference at the hot outlet end",
        )
        if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
            self.crossflow_factor = Var(
                self.flowsheet().config.time,
                initialize=1.0,
                doc="Factor to adjust coutercurrent flow heat "
                "transfer calculation for cross flow.",
            )
            f = self.crossflow_factor
        ########################################################################
        # Add control volumes                                                  #
        ########################################################################
        _make_heater_control_volume(
            self,
            "side_1",
            config.hot_side_config,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        _make_heater_control_volume(
            self,
            "side_2",
            config.cold_side_config,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        # Add named references to side_1 and side_2, side 1 and 2 maintain
        # backward compatability and are names the user doesn't need to worry
        # about. The sign convention for duty is heat from side 1 to side 2 is
        # positive
        add_object_reference(self, config.hot_side_name, self.side_1)
        add_object_reference(self, config.cold_side_name, self.side_2)

        # Add convienient references to heat duty.
        q = self.heat_duty = Reference(self.side_2.heat)
        ########################################################################
        # Add ports                                                            #
        ########################################################################
        # Keep old port names, just for backward compatability
        self.add_inlet_port(name="inlet_1", block=self.side_1, doc="Hot side inlet")
        self.add_inlet_port(name="inlet_2", block=self.side_2, doc="Cold side inlet")
        self.add_outlet_port(name="outlet_1", block=self.side_1, doc="Hot side outlet")
        self.add_outlet_port(name="outlet_2", block=self.side_2, doc="Cold side outlet")

        # Using Andrew's function for now, I think Pyomo's refrence has trouble
        # with scalar (pyomo) components.
        add_object_reference(self, config.hot_side_name + "_inlet", self.inlet_1)
        add_object_reference(self, config.cold_side_name + "_inlet", self.inlet_2)
        add_object_reference(self, config.hot_side_name + "_outlet", self.outlet_1)
        add_object_reference(self, config.cold_side_name + "_outlet", self.outlet_2)
        ########################################################################
        # Add end temperature differnece constraints                            #
        ########################################################################
        @self.Constraint(self.flowsheet().config.time)
        def delta_temperature_in_equation(b, t):
            if b.config.flow_pattern == HeatExchangerFlowPattern.cocurrent:
                return (
                    b.delta_temperature_in[t]
                    == b.side_1.properties_in[t].temperature
                    - b.side_2.properties_in[t].temperature
                )
            else:
                return (
                    b.delta_temperature_in[t]
                    == b.side_1.properties_in[t].temperature
                    - b.side_2.properties_out[t].temperature
                )

        @self.Constraint(self.flowsheet().config.time)
        def delta_temperature_out_equation(b, t):
            if b.config.flow_pattern == HeatExchangerFlowPattern.cocurrent:
                return (
                    b.delta_temperature_out[t]
                    == b.side_1.properties_out[t].temperature
                    - b.side_2.properties_out[t].temperature
                )
            else:
                return (
                    b.delta_temperature_out[t]
                    == b.side_1.properties_out[t].temperature
                    - b.side_2.properties_in[t].temperature
                )

        ########################################################################
        # Add a unit level energy balance                                      #
        ########################################################################
        @self.Constraint(self.flowsheet().config.time)
        def unit_heat_balance(b, t):
            return 0 == self.side_1.heat[t] + self.side_2.heat[t]

        ########################################################################
        # Add delta T calculations using callack function, lots of options,    #
        #   and users can provide their own if needed                          #
        ########################################################################
        config.delta_temperature_callback(self)
        ########################################################################
        # Add Heat transfer equation                                           #
        ########################################################################
        deltaT = self.delta_temperature
        scale = self.side_1.scaling_factor_energy

        @self.Constraint(self.flowsheet().config.time)
        def heat_transfer_equation(b, t):
            if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
                return 0 == (f[t] * u[t] * a * deltaT[t] - q[t]) * scale
            else:
                return 0 == (u[t] * a * deltaT[t] - q[t]) * scale

        ########################################################################
        # Add symbols for LaTeX equation rendering                             #
        ########################################################################
        self.overall_heat_transfer_coefficient.latex_symbol = "U"
        self.area.latex_symbol = "A"
        self.side_1.heat.latex_symbol = "Q_1"
        self.side_2.heat.latex_symbol = "Q_2"
        self.delta_temperature.latex_symbol = "\\Delta T"

    def initialize(
        self,
        state_args_1=None,
        state_args_2=None,
        outlvl=0,
        solver="ipopt",
        optarg={"tol": 1e-6},
        duty=1000,
    ):
        """
        Heat exchanger initialization method.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for side_1 (see documentation of the specific
                property package) (default = {}).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for side_2 (see documentation of the specific
                property package) (default = {}).
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating which solver to use during
                     initialization (default = 'ipopt')
            duty : an initial guess for the amount of heat transfered
                (default = 10000)

        Returns:
            None

        """
        # Set solver options
        init_log = getInitLogger(self.name, outlvl)
        opt = SolverFactory(solver)
        opt.options = optarg
        flags1 = self.side_1.initialize(
            outlvl=outlvl + 1, optarg=optarg, solver=solver, state_args=state_args_1
        )

        init_log.log(4, "Initialization Step 1a (side_1) Complete.")

        flags2 = self.side_2.initialize(
            outlvl=outlvl + 1, optarg=optarg, solver=solver, state_args=state_args_2
        )

        init_log.log(4, "Initialization Step 1b (side_2) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.heat_transfer_equation.deactivate()
        self.side_2.heat.fix(duty)
        results = opt.solve(self, tee=init_tee(init_log), symbolic_solver_labels=True)
        init_log.log(
            4, "Initialization Step 2 {}.".format(results.solver.termination_condition)
        )
        self.side_2.heat.unfix()
        self.heat_transfer_equation.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        results = opt.solve(self, tee=init_tee(init_log), symbolic_solver_labels=True)
        init_log.log(4, "Initialization Step 3 {}.".format(condition(results)))
        # ---------------------------------------------------------------------
        # Release Inlet state
        self.side_1.release_state(flags1, outlvl + 1)
        self.side_2.release_state(flags2, outlvl + 1)

        init_log.log(5, "Initialization Completed, {}".format(condition(results)))

    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "HX Coefficient": self.overall_heat_transfer_coefficient[time_point]
        }
        var_dict["HX Area"] = self.area
        var_dict["Heat Duty"] = self.heat_duty[time_point]
        if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
            var_dict = {"Crossflow Factor": self.crossflow_factor[time_point]}

        expr_dict = {}
        expr_dict["Delta T Driving"] = self.delta_temperature[time_point]
        expr_dict["Delta T In"] = self.delta_temperature_in[time_point]
        expr_dict["Delta T Out"] = self.delta_temperature_out[time_point]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot Inlet": self.inlet_1,
                "Hot Outlet": self.outlet_1,
                "Cold Inlet": self.inlet_2,
                "Cold Outlet": self.outlet_2,
            },
            time_point=time_point,
        )

    def get_costing(self, module=costing):
        if not hasattr(self.flowsheet(), "costing"):
            self.flowsheet().get_costing()
        self.costing = Block()

        module.hx_costing(self.costing)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Plot doalamo output including confidence intervals if they are calculated.
"""


def almplot(res, show=True):
    try:
        import matplotlib.pyplot as plt
        import numpy as np
        from alamopy.writethis import writethis
    except Exception:
        writethis("Cannot plot, possibly missing matplotlib package")

    model = res['model'].replace(' - ', ' + -') 
    model = model.split('=')[1]
    model = model.split(' + ')
    if model[0] == ' ':  # if there are more than one terms, the first split is ' '
        model = model[1:]
    ndp = 100
    t = np.linspace(0.08, 1.7, ndp)
    out = np.zeros([3, ndp])
    clo = np.zeros(ndp)
    chi = np.zeros(ndp)
    coeff = np.zeros(ndp)
    
    for i in range(len(model)):
        coeff[i] = float(model[i].split(' * ')[0])
        if 'conf_inv' in res.keys():
            clo[i] = coeff[i] - float(res['conf_inv'][i].split('+/-')[1])
            chi[i] = coeff[i] + float(res['conf_inv'][i].split('+/-')[1])
        
    for i in range(ndp):
        out[0, i] = float(coeff[0]) * t[i]**1.2 \
            - float(coeff[1]) * t[i]**2 - float(coeff[2])
        if 'conf_inv' in res.keys():  # If confidence intervals exist
            out[1, i] = clo[0] * t[i]**1.2 - chi[1] * t[i]**2 - chi[2]
            out[2, i] = chi[0] * t[i]**1.2 - clo[1] * t[i]**2 - clo[2]

    plt.plot(t, out[0], 'b-')
    if 'conf_inv' in res.keys():
        plt.plot(t, out[1], 'r--', t, out[2], 'r--')
    if show:
        plt.show()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
from idaes.surrogate import ripe
import numpy as np


def main():
    spec = ['X']
    # Import data from csv
    data = np.genfromtxt('clc.csv', delimiter=',')
    t = data[:,0]
    xdata = data[:,1]
    stoich = [1]

    # User pre-defined clc rate forms found in RIPE
    mechs = ripe.clcforms

    # Identify optimal kinetic mechanism
    results = ripe.ripemodel(xdata,stoichiometry=stoich,mechanisms=mechs,time=t)


if __name__ == "__main__":
    main()


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Resource database.
"""
# system
from datetime import datetime
import logging
import re

# third party
import pendulum
from tinydb import TinyDB, Query

# local
from . import errors
from .resource import Resource
from .resource import Triple, triple_from_resource_relations

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


class ResourceDB(object):
    """A database interface to all the resources within a given DMF workspace.
    """

    def __init__(self, dbfile=None, connection=None):
        """Initialize from DMF and given configuration field.

        Args:
            dbfile (str): DB location
            connection: If non-empty, this is an
                existing connection that should be re-used, instead of
                trying to connect to the location in `dbfile`.

        Raises:
            ValueError, if dbfile
             and connection are both None
        """
        self._db = None
        self._gr = None

        if connection is not None:
            self._db = connection
        elif dbfile is not None:
            try:
                db = TinyDB(dbfile)
            except IOError:
                raise errors.FileError('Cannot open resource DB "{}"'.format(dbfile))
            # turn off caching, otherwise update() does not work properly
            self._db = db.table('resources', cache_size=0)

    def __len__(self):
        return len(self._db)

    def find(self, filter_dict, id_only=False, flags=0):
        """Find and return records based on the provided filter.

        Args:
            filter_dict (dict): Search filter. For syntax, see docs in
                                :meth:`.dmf.DMF.find`.
            id_only (bool): If true, return only the identifier of each
                resource; otherwise a Resource object is returned.
            flags (int): Flag values for, e.g., regex searches

        Returns:
            generator of int|Resource, depending on the value of `id_only`
        """

        def as_resource(_r):
            _log.debug(f"as_resource: id_={_r['id_']}")
            rsrc = Resource(value=_r)
            rsrc.v['doc_id'] = _r.doc_id
            return rsrc

        # with no filter, do a find-all
        if not filter_dict:
            for r in self._db.all():
                if id_only:
                    yield r.eid
                else:
                    yield as_resource(r)
            return
        filter_expr = self._create_filter_expr(filter_dict, flags)

        # return results for query
        _log.debug('Find resources matching: {}'.format(filter_expr))
        results = self._db.search(filter_expr)
        for r in results:
            if id_only:
                yield r.eid
            else:
                _log.debug(f"got resource: {r}")
                yield as_resource(r)

    @classmethod
    def _create_filter_expr(cls, filter_dict, flags=0):
        # convert filter to expression for TinyDB
        q = Query()
        filter_expr = None

        def fadd(c):
            """Update filter, or set to 1st condition, & return new value."""
            return c if filter_expr is None else filter_expr & c

        for k, v in filter_dict.items():
            if not k:
                continue  # XXX: Issue a warning?
            # strip off list-query operator
            qry_all = False
            if isinstance(v, list) and k.endswith('!'):
                k, qry_all = k[:-1], True
            # Get the query[field1][field2][..][fieldN] object
            qry = q
            for field in k.split('.'):
                qry = qry[field]
            # A list value means find these value(s) in the
            # list value of the record. By default, any matching
            # value is a match. If the key has a '!' appended, then
            # require all values for a match.
            if isinstance(v, list):
                if len(v) > 0:
                    # if values are dicts, nest a query
                    # XXX: this only works one level deep
                    # XXX: only one nested query is used/allowed
                    if isinstance(v[0], dict):
                        list_expr = None
                        for list_k, list_v in v[0].items():
                            list_qry = Query()
                            for field in list_k.split('.'):
                                list_qry = list_qry[field]
                            list_cond = cls._expr_to_query(
                                list_qry, list_v, flags=flags
                            )
                            if list_expr is None:
                                list_expr = list_cond
                            else:
                                list_expr = list_expr & list_cond
                    # otherwise, simply put the values in there
                    else:
                        list_expr = tuple(v)
                    # add nested "query" (or value tuple)
                    if qry_all:
                        cond = qry.all(list_expr)
                    else:
                        cond = qry.any(list_expr)
            else:
                cond = cls._expr_to_query(qry, v, flags=flags)
            filter_expr = fadd(cond)
        return filter_expr

    @classmethod
    def _expr_to_query(cls, qry, v, flags=None):
        """Get a query from a filter expr.

        There are two types of non-list values:
         1. equality is just {key: value}
         2. inequalities are {key: {op: value, ...}}
           operators are "$<shell test op>", such as "$lt",
           just like in MongoDB; see _op_cond() method for details.
        """
        result = None
        if isinstance(v, dict):
            for op_key, op_value in v.items():
                tv = cls._value_transform(op_value)
                cond = cls._op_cond(qry, op_key, tv)
                result = cond if result is None else result & cond
        else:
            tv = cls._value_transform(v)
            if v is True:
                result = qry.exists()
            elif tv is False:
                result = ~qry.exists()
            elif hasattr(tv, "match"):  # regex
                result = qry.matches(tv.pattern, flags=flags)
            else:
                result = qry == tv
        return result

    @staticmethod
    def _value_transform(v):
        # transform dates into timestamps
        if isinstance(v, datetime) or isinstance(v, pendulum.Pendulum):
            if isinstance(v, datetime):
                pv = pendulum.create(
                    v.year,
                    v.month,
                    v.day,
                    v.hour,
                    v.minute,
                    v.second,
                    v.microsecond,
                    v.tzname(),
                )
            else:
                pv = v
            return pv.timestamp()
        # support for special string '@' values
        elif isinstance(v, str) and len(v) > 0 and v[0] == '@':
            if v == '@true':
                return True
            if v == '@false':
                return False
            return v
        # for a regex, return a Python regex obj
        elif isinstance(v, str) and len(v) > 0 and v[0] == '~':
            return re.compile(v[1:])
        # default is no transformation
        else:
            return v

    @staticmethod
    def _op_cond(query, op, value):
        # sanity check that operator is truthy
        if not op:
            raise ValueError(f"empty operator for value `{value}`")
        # just a clumsy switch statement..
        if op == '$gt':
            cond = query > value
        elif op == '$ge':
            cond = query >= value
        elif op == '$lt':
            cond = query < value
        elif op == '$le':
            cond = query <= value
        elif op == '$ne':
            cond = query != value
        else:
            raise ValueError('Unexpected operator: {}'.format(op))
        return cond

    def find_one(self, *args, **kwargs):
        """Same as `find()`, but returning only first value or None.
        """
        result = None
        for value in self.find(*args, **kwargs):
            result = value
            break
        return result

    def find_related(self, id_, filter_dict=None, outgoing=True, maxdepth=0, meta=None):
        """Find all resources connected to the identified one.

        Args:
            id_ (str): Unique ID of target resource.
            filter_dict (dict): Filter to these resources
            outgoing:
            maxdepth:
            meta (List[str]): Metadata fields to extract
        Returns:
            Generator of (depth, relation, metadata)
        Raises:
            KeyError if the resource is not found.
        """
        if maxdepth <= 0:
            maxdepth = 9223372036854775807
        # Get an iterator over all the resources, optionally
        # filtered by an expression, as for find().
        if filter_dict:
            filter_expr = self._create_filter_expr(filter_dict)
            resource_list = self._db.search(filter_expr)
        else:
            resource_list = self._db.all()
        # build adjacency list representing connections between resources
        relation_map = {}
        for rsrc in resource_list:
            for rrel in rsrc['relations']:
                uuid = rsrc[Resource.ID_FIELD]
                rel = triple_from_resource_relations(uuid, rrel)
                # skip start of edge, get metadata, etc. at end of edge
                if (outgoing and rel.subject == uuid) or (
                    not outgoing and rel.object == uuid
                ):
                    continue
                # add entry in map
                key = rel.subject if outgoing else rel.object
                meta_info = {k: rsrc[k] for k in meta}
                value = (rel.subject, rel.predicate, rel.object, meta_info)
                value_list = relation_map.get(key, None)
                if value_list is None:
                    relation_map[key] = [value]
                else:
                    value_list.append(value)
        _log.debug(f"built relation map: {relation_map}")
        # stop if there are no connections
        if id_ not in relation_map:
            return
        # Do a depth-first search through the edges, yield-ing
        # the relations as we go
        q, depth, visited = [], 0, set()
        q.extend(relation_map[id_])
        visited.add(id_)
        while len(q) > 0 and depth < maxdepth:
            depth += 1
            # visit all the nodes in the queue
            n = len(q)
            for i in range(n):
                relation = Triple(*q[i][:3])
                yield (depth, relation, q[i][3])
                if depth < maxdepth:
                    # Follow relations from subject or object, depending on
                    # the "direction" that we are searching.
                    next_id = relation.object if outgoing else relation.subject
                    # If there are relations, and we haven't already been to
                    # this node, add them at the end of the queue; we will
                    # visit them at the next depth increment.
                    if next_id in relation_map and next_id not in visited:
                        q.extend(relation_map[next_id])
                        visited.add(next_id)
            q = q[n:]  # pop off all the nodes we just visited

    def get(self, identifier):
        """Get a resource by identifier.

        Args:
          identifier: Internal identifier

        Returns:
            (Resource) A resource or None
        """

        def as_resource(_r):
            rsrc = Resource(value=_r)
            rsrc.v['doc_id'] = _r.doc_id
            return rsrc

        item = self._db.get(doc_id=identifier)
        if item is None:
            return None
        return as_resource(item)

    def put(self, resource):
        """Put this resource into the database.

        Args:
            resource (Resource): The resource to add

        Returns:
            None

        Raises:
            errors.DuplicateResourceError: If there is already a resource
                in the database with the same "id".
        """
        _log.debug(f"put resource id={resource.id}")
        # check for same id
        qry = Query()
        if self._db.contains(qry.id_ == resource.id):
            raise errors.DuplicateResourceError("put", resource.id)
        # add resource
        self._db.insert(resource.v)

    def delete(self, id_=None, idlist=None, filter_dict=None, internal_ids=False):
        """Delete one or more resources with given identifiers.

        Args:
            id_ (Union[str,int]): If given, delete this id.
            idlist (list): If given, delete ids in this list
            filter_dict (dict): If given, perform a search and
                           delete ids it finds.
            internal_ids (bool): If True, treat identifiers as numeric
                (internal) identifiers. Otherwise treat them as
                resource (string) indentifiers.
        Returns:
            (list[str]) Identifiers
        """
        if internal_ids:
            doc_ids = idlist if idlist else [id_]
            self._db.remove(doc_ids=doc_ids)
        else:
            ID = Resource.ID_FIELD
            if filter_dict:
                cond = self._create_filter_expr(filter_dict)
            elif id_:
                cond = self._create_filter_expr({ID: id_})
            elif idlist:
                cond = self._create_filter_expr({ID: [idlist]})
            else:
                return
            self._db.remove(cond=cond)

    def update(self, id_, new_dict):
        """Update the identified resource with new values.

        Args:
            id_ (int): Identifier of resource to update
            new_dict (dict): New dictionary of resource values
        Returns:
            None
        Raises:
            ValueError: If new resource is of wrong type
            KeyError: If old resource is not found
        """
        _log.debug("update.start")
        id_cond = {Resource.ID_FIELD: id_}
        old = self.find_one(id_cond)
        if old is None:
            raise KeyError('Cannot find resource id={}'.format(id_))
        T = Resource.TYPE_FIELD
        if old.v[T] != new_dict[T]:
            raise ValueError(
                'New resource type="{}" does not '
                'match current resource type "{}"'.format(new_dict[T], old.v[T])
            )
        _log.debug(f"update old resource: {old.v}")
        changed = {}
        for k, v in new_dict.items():
            if k not in old.v:
                changed[k] = v
            elif old.v[k] != v:
                changed[k] = v
        _log.debug(f"update resource {id_} with new values: {changed}")
        self._db.update(changed, self._create_filter_expr(id_cond))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This file contains a reactor simulator built in pyomo
# The simulator emulates behavior observed in the production of
# Styrene from ethylbenzene

import numpy as np
import pyomo.environ as pyo

# define fractional variance of noise, SNR = 1 / noise
noise = 0.05

# Reference temperature defined
Tr = 750.0
# Kinetic parameters are hard coded in the form [[k1,k2...],[E1,E2,...]]
kinetic_params = [[250,220,38,25],[115,131,55,75]]


def sim(data):
    import numpy as np
    # Enable 1/2d calls
    # Ensure that data sizes and shapes are consistent
    try:
        x0 = data[:,:9]
        Temp = data[:,9]
    except:
        x0 = data[:9]
        Temp = data[9]
    params = kinetic_params
    dshape = np.shape(x0)
    if len(dshape) == 1:
        x0 = np.expand_dims(x0, axis=-1)
        x0 = np.ndarray.transpose(x0)
        dshape = np.shape(x0)
    npc = dshape[0]
    ns = dshape[1]
    # Define the reactor simulation in pyomo
    def pyomosim(data):
        # Define rate parameters 
        kco = 35
        kst = 1.5
        sparam = 1.0
        flow = 1.0
        vol = 1.0
        gc = .008314
        # Define kinetic rate parameters
        k = params[0]
        E = params[1]
        # define UB for concentration
        #nound_ub = 20
        # pyomo solver options
        opt = pyo.SolverFactory('baron')
        cracmodel = pyo.ConcreteModel()
        ca0,cb0,cc0,cd0,cf0,cg0,ch0,ci0,cj0,T = [float(v) for v in data]
        
        bound_ub = 100.0
        # Define cracmodel variables
        # A = Eb , B = St , C = Bz , D = Et, E = Tl, F = Me, G = Water, H = H2 I = CO2, J = N2
        cracmodel.a = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = ca0)
        cracmodel.b = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cb0)
        cracmodel.c = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cc0)
        cracmodel.d = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cd0)
        cracmodel.f = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cf0)
        cracmodel.g = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cg0)
        cracmodel.h = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = ch0)
        cracmodel.i = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = ci0)
        cracmodel.j = pyo.Var(domain = pyo.NonNegativeReals, bounds = (0,bound_ub), initialize = cj0)
        cracmodel.r1 = pyo.Var(domain = pyo.Reals)
        cracmodel.r2 = pyo.Var(domain = pyo.Reals)
        cracmodel.r4 = pyo.Var(domain = pyo.Reals)
        cracmodel.r5 = pyo.Var(domain = pyo.Reals)

        def keq(T):
            return pyo.exp(0.1 + (300 / T))
#            return pyo.exp(-1.0*(122700-126.3*T-0.002194*T**2)/(gc*T)) * sparam
        # define reaction rate variable
        
        def fr1(cracmodel):
            # A <> B + H
            return cracmodel.r1 == k[0] * pyo.exp(-(E[0]/(gc))*((1/T)-(1/Tr))) * (cracmodel.a - (cracmodel.b * cracmodel.h)/keq(T)) * (1.0/((1+kst*cracmodel.b)*(1+kco*cracmodel.i)))
        def fr2(cracmodel):
            # A > C + D
            return cracmodel.r2 == k[1] * pyo.exp(-(E[1]/(gc))*((1/T)-(1/Tr)))  * cracmodel.a * (1.0/(1+kco*cracmodel.i))

        def fr4(cracmodel):
            # D + 2H > 2F
            return cracmodel.r4 == k[2] * pyo.exp(-(E[2]/(gc))*((1/T)-(1/Tr)))  * cracmodel.d * cracmodel.h
        def fr5(cracmodel):
            # F+G > I + 4H
            return cracmodel.r5 == k[3] * pyo.exp(-(E[3]/(gc))*((1/T)-(1/Tr)))  * cracmodel.f * cracmodel.g

        cracmodel.er1 = pyo.Constraint( rule = fr1)
        cracmodel.er2 = pyo.Constraint( rule = fr2)
        cracmodel.er4 = pyo.Constraint( rule = fr4)
        cracmodel.er5 = pyo.Constraint( rule = fr5)
        cracmodel.sets = pyo.RangeSet(9)
        cracmodel.dum = pyo.Var(cracmodel.sets, domain = pyo.Reals)

        def fra(cracmodel):  # A - 1,2,3
            return cracmodel.dum[1]  == ca0-cracmodel.a - cracmodel.r1 -cracmodel.r2 
        def frb(cracmodel): # B - 1
            return cracmodel.dum[2]  == cb0-cracmodel.b+cracmodel.r1
        def frc(cracmodel): # C - 2
            return cracmodel.dum[3]  == cc0-cracmodel.c+cracmodel.r2
        def frd(cracmodel): # D - 2,4
            return cracmodel.dum[4]  == cd0-cracmodel.d+cracmodel.r2-cracmodel.r4
        def frf(cracmodel): # F - 3,4,5
            return cracmodel.dum[5]  == cf0-cracmodel.f+2*cracmodel.r4-cracmodel.r5
        def frg(cracmodel): # G - 5
            return cracmodel.dum[6]  == cg0-cracmodel.g-2*cracmodel.r5
        def frh(cracmodel): # H - 1,3,4,5
            return cracmodel.dum[7]  == ch0-cracmodel.h+cracmodel.r1-2*cracmodel.r4+4*cracmodel.r5
        def fri(cracmodel): # I - 5
            return cracmodel.dum[8]  == ci0-cracmodel.i + cracmodel.r5
        def frj(cracmodel): # J - N2 is inert
            return cracmodel.dum[9] == cj0-cracmodel.j

        cracmodel.era = pyo.Constraint( rule = fra)
        cracmodel.erb = pyo.Constraint( rule = frb)
        cracmodel.erc = pyo.Constraint( rule = frc)
        cracmodel.erd = pyo.Constraint( rule = frd)
        cracmodel.erf = pyo.Constraint( rule = frf)
        cracmodel.erg = pyo.Constraint( rule = frg)
        cracmodel.erh = pyo.Constraint( rule = frh)
        cracmodel.eri = pyo.Constraint( rule = fri)
        cracmodel.erj = pyo.Constraint( rule = frj)

        # minimize square of dummy variables to find steady-state concentrations
        def objf(cracmodel):
            return sum(cracmodel.dum[s]**2 for s in cracmodel.sets)
        
        cracmodel.OBJ = pyo.Objective(rule = objf)
        results = opt.solve(cracmodel)
        cracmodel.solutions.store_to(results)
        klist = ['a','b','c','d','f','g','h','i','j']
        # Add noise of the specifiec SNR, noise has variance eps ~ N(0,noise*conc)
        vn = [results.Solution.Variable[key]['Value']+np.random.normal(0,noise*results.Solution.Variable[key]['Value']) for key in klist]
        return vn
    
    # Simulate over requested datapoints
    # requested data may have 1 or more points
    concentrations = []
    if npc != 1:
        for i in range(npc):
            try:
                t2 = Temp[0][i]
            except:
                t2 = Temp[i]            
            conres = pyomosim(np.ndarray.tolist(x0[i,:])+[t2])

            concentrations.append(conres)
    else:
        conres = pyomosim(data)
        concentrations = conres
    return concentrations



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

from .multos import deletefile, movefile, catfile


def almwriter(data, debug, vargs, kwargs):
    # This function writes a .alm file for the given data

    xdata = vargs[0]
    zdata = vargs[1]
    if (debug['validation']):
        xvaldata = vargs[2]
        zvaldata = vargs[3]

    with open(data['stropts']['almname'], 'w') as a:
        for arg in data['opts'].keys():
            if arg == 'sigma' and data['opts'][arg] < 0 : 
                continue
            elif (arg == 'modeler' and data['opts'][arg] == 6):
                a.write(arg + ' ' + str(data['opts'][arg]) + '\n')
                a.write('builder 0\n')
                a.write('enumall 1\n')
                a.write('enum1 1\n')          
            else: 
                a.write(arg + ' ' + str(data['opts'][arg]) + '\n')

        for arg in list(['xlabels', 'zlabels']):  # KAIWEN
            a.write(arg + ' ' + str(data['labs']['save' + arg])[1:-1]
                    .replace(",", '').replace("'", '') + '\n')
        # for arg in list(['xlabels','zlabels']):
        # a.write(arg+' '+str(data['lstopts'][arg])[1:-1]
        # .replace(",",'').replace("'",'')+'\n')
        
        for arg in data['lstopts'].keys():
            if arg in kwargs.keys() and arg not in list(['xlabels', 'zlabels']):
                temp = arg + ' '
                for value in data['lstopts'][arg]:
                    value = ''.join(c for c in str(value) if c not in '(),')
                    temp = temp + str(value) + ' '
                    temp = temp.replace('[', '')
                    temp = temp.replace(']', '')
                a.write(temp + '\n')

        a.write('trace 1\n')
        for arg in data['stropts'].keys():
            if (arg not in list(['almopt', 'almname'])):
                a.write(arg + ' ' + str(data['stropts'][arg]) + '\n')

        # xmax and xmin writing in set4
        if ('xmin' in kwargs.keys()):
            for arg in data['set4'].keys():
                temp = arg + ' '
                for value in data['set4'][arg]:
                    value = ''.join(c for c in str(value) if c not in "(),[] ")
                    temp = temp + str(value) + ' '
                a.write(temp + '\n')
        else:
            for arg in data['set4'].keys():
                a.write(arg + ' ' + str(data['set4'][arg])[0:-1]
                        .replace(",", '') + '\n')

        # Write data and valdata
        a.write('begin_data\n')  # ISSUE ENGLE
        for i in range(data['opts']['ndata']):
            temp = ''
            if data['opts']['ninputs'] > 1:
                for j in range(data['opts']['ninputs']):
                    temp = temp + str(xdata[i][j]) + ' '
            else:
                temp = temp + str(xdata[i]) + ' '
            for j in range(data['opts']['noutputs']):
                temp = temp + str(zdata[i][j]) + ' '
            a.write(temp + '\n')
        a.write('end_data\n')
        if(debug['validation']):
            a.write('begin_valdata\n')
            for i in range(data['opts']['nvaldata']):
                temp = ''
                for j in range(data['opts']['ninputs']):
                    temp = temp + str(xvaldata[i][j]) + ' '
                for j in range(data['opts']['noutputs']):
                    temp = temp + str(zvaldata[i][j]) + ' '
                a.write(temp + '\n')
            a.write('end_valdata\n')
        
        if 'solvemip' in data['opts'].keys():
            if data['opts']['solvemip'] == 1:
                a.write('gams ' + debug['gamsloc'] + '\n')
        if debug['savegams']:
            a.write('enum1 0\n')  
            a.write('enumall 0\n')    
            a.write('removescratch 0\n')    
        # make modeler=6 OLS
        # a.write('cvxbic 1\n')
        # a.write('convpen 0\n')

        # This option can be used to test the flow of alamo
        if (debug['hardset']):
            temp = 'maxterms '
            for i in range(data['opts']['noutputs']):
                temp = temp + '1 '
            a.write(temp)

    # Append text file if specified
    if ('almopt' in data['stropts'].keys()):

        # Kaiwen
        # destination = open('almtemp','wb')
        # shutil.copyfileobj(open(str(data['stropts']['almname']),'rb'), destination)
        # shutil.copyfileobj(open(str(data['stropts']['almopt']),'rb'), destination)
        # destination.close()
        # os.remove(str(data['stropts']['almname']))
        # os.remove(str(data['stropts']['almopt']))
        # os.rename('almtemp',str(data['stropts']['almname']))

        catfile('almtemp', str(data['stropts']['almname']),
                str(data['stropts']['almopt']))
        deletefile(str(data['stropts']['almname']))
        movefile('almtemp ' + str(data['stropts']['almname']))
